from aerosandbox.common import AeroSandboxObject
from aerosandbox.geometry.common import *
from typing import List, Dict, Any, Tuple, Union, Optional, Callable
from aerosandbox.geometry.airfoil import Airfoil
from numpy import pi
import aerosandbox.numpy as np
import aerosandbox.geometry.mesh_utilities as mesh_utils
import copy


class Wing(AeroSandboxObject):
    """
    Definition for a Wing.

    Anatomy of a Wing:

        A wing consists chiefly of a collection of cross-sections, or "xsecs". A cross-section is a 2D "slice" of a
        wing. These can be accessed with `Wing.xsecs`, which gives a list of xsecs in the Wing. Each xsec is a
        WingXSec object, a class that is defined separately.

        You may also see references to wing "sections", which are different than cross-sections (xsecs)! Sections are
        the portions of the wing that are in between xsecs. In other words, a wing with N cross-sections (xsecs,
        WingXSec objects) will always have N-1 sections. Sections are never explicitly defined, since you can get all
        needed information by lofting from the adjacent cross-sections. For example, section 0 (the first one) is a
        loft between cross-sections 0 and 1.

        Wings are lofted linearly between cross-sections.

    If the wing is symmetric across the XZ plane, just define the right half and supply `symmetric=True` in the
    constructor.

    If the wing is not symmetric across the XZ plane (e.g., a single vertical stabilizer), just define the wing.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 xsecs: List['WingXSec'] = None,
                 symmetric: bool = False,
                 color: Optional[Union[str, Tuple[float]]] = None,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 **kwargs,  # Only to allow for capturing of deprecated arguments, don't use this.
                 ):
        """
        Defines a new wing object.

        Args:

            name: Name of the wing [optional]. It can help when debugging to give each wing a sensible name.

            xsecs: A list of wing cross-sections ("xsecs") in the form of WingXSec objects.

            symmetric: Is the wing symmetric across the XZ plane?

            color: Determines what color to use for this component when drawing the airplane. Optional,
                and for visualization purposes only. If left as None, a default color will be chosen at the time of
                drawing (usually, black). Can be any color format recognized by MatPlotLib, namely:

                * A RGB or RGBA tuple of floats in the interval [0, 1], e.g., (0.1, 0.2, 0.5, 0.3)

                * Case-insensitive hex RGB or RGBA string, e.g., '#0f0f0f80'

                * String representation of float value in closed interval [0, 1] for grayscale values, e.g.,
                    '0.8' for light gray

                * Single character shorthand notation for basic colors, e.g., 'k' -> black, 'r' -> red

                See also: https://matplotlib.org/stable/tutorials/colors/colors.html

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
                that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }
        """
        ### Set defaults
        if name is None:
            name = "Untitled"
        if xsecs is None:
            xsecs: List['WingXSec'] = []
        if analysis_specific_options is None:
            analysis_specific_options = {}

        ### Initialize
        self.name = name
        self.xsecs = xsecs
        self.symmetric = symmetric
        self.color = color
        self.analysis_specific_options = analysis_specific_options

        ### Handle deprecated parameters
        if 'xyz_le' in locals():
            import warnings
            warnings.warn(
                "The `xyz_le` input for Wing is pending deprecation and will be removed in a future version. Use Wing().translate(xyz) instead.",
                stacklevel=2
            )
            self.xsecs = [
                xsec.translate(xyz_le)
                for xsec in self.xsecs
            ]

    def __repr__(self) -> str:
        n_xsecs = len(self.xsecs)
        symmetry_description = "symmetric" if self.symmetric else "asymmetric"
        return f"Wing '{self.name}' ({len(self.xsecs)} {'xsec' if n_xsecs == 1 else 'xsecs'}, {symmetry_description})"

    def translate(self,
                  xyz: Union[np.ndarray, List[float]]
                  ) -> 'Wing':
        """
        Translates the entire Wing by a certain amount.

        Args:
            xyz:

        Returns: The new wing object.

        """
        new_wing = copy.copy(self)
        new_wing.xsecs = [
            xsec.translate(xyz)
            for xsec in new_wing.xsecs
        ]
        return new_wing

    def span(self,
             type: str = "yz",
             include_centerline_distance=False,
             _sectional: bool = False,
             ) -> Union[float, List[float]]:
        """
        Computes the span, with options for various ways of measuring this (see `type` argument).

        If the wing is symmetric, both left/right sides are included in order to obtain the full span. In the case
        where the root cross-section is not coincident with the center plane (e.g., XZ plane), this function's
        behavior depends on the `include_centerline_distance` argument.

        Args:

            type: One of the following options, as a string:

                * "xyz": First, computes the quarter-chord point of each WingXSec. Then, connects these with
                straight lines. Then, adds up the lengths of these lines.

                * "xy" or "top": Same as "xyz", except it projects each line segment onto the XY plane before adding up the
                lengths.

                * "yz" (default) or "front": Same as "xyz", except it projects each line segment onto the YZ plane (i.e., front view)
                before adding up the lengths.

                * "xz" or "side": Same as "xyz", except it projects each line segment onto the XZ plane before adding up the
                lengths. Rarely needed.

                * "x": Same as "xyz", except it only counts the x-components of each line segment when adding up the
                lengths.

                * "y": Same as "xyz", except it only counts the y-components of each line segment when adding up the
                lengths.

                * "z": Same as "xyz", except it only counts the z-components of each line segment when adding up the
                lengths.

            include_centerline_distance: A boolean flag that tells the function what to do if a wing's root is not
            coincident with the centerline plane (i.e., XZ plane).

                * If True, we first figure out which WingXSec has its quarter-chord point closest to the centerline
                plane (i.e., XZ plane). Then, we compute the distance from that quarter-chord point directly to the
                centerline plane (along Y). We then add that distance to the span calculation. In other words,
                the fictitious span connecting the left and right root cross-sections is included.

                * If False, this distance is ignored. In other words, the fictitious span connecting the left and
                right root cross-sections is not included. This is the default behavior.

                Note: For computation, either the root WingXSec (i.e., index=0) or the tip WingXSec (i.e., index=-1)
                is used, whichever is closer to the centerline plane. This will almost-always be the root WingXSec,
                but some weird edge cases (e.g., a half-wing defined on the left-hand-side of the airplane,
                rather than the conventional right-hand side) will result in the tip WingXSec being used.

            _sectional: A boolean. If False, returns the total span. If True, returns a list of spans for each of the
                `n-1` lofted sections (between the `n` wing cross-sections in wing.xsec).
        """
        # Check inputs
        if include_centerline_distance and _sectional:
            raise ValueError("Cannot use `_sectional` with `include_centerline_distance`!")

        # Handle overloaded names
        if type == "top":
            type = "xy"
        elif type == "front":
            type = "yz"
        elif type == "side":
            type = "xz"

        # Figure out where the quarter-chord points of each WingXSec are
        i_range = range(len(self.xsecs))

        quarter_chord_locations = [
            self._compute_xyz_of_WingXSec(
                i,
                x_nondim=0.25,
                z_nondim=0,
            )
            for i in i_range
        ]

        # Compute sectional spans
        sectional_spans = []

        for inner_i, outer_i in zip(i_range, i_range[1:]):
            quarter_chord_vector = (
                    quarter_chord_locations[outer_i] -
                    quarter_chord_locations[inner_i]
            )

            if type == "xyz":
                section_span = (
                                       quarter_chord_vector[0] ** 2 +
                                       quarter_chord_vector[1] ** 2 +
                                       quarter_chord_vector[2] ** 2
                               ) ** 0.5

            elif type == "xy":
                section_span = (
                                       quarter_chord_vector[0] ** 2 +
                                       quarter_chord_vector[1] ** 2
                               ) ** 0.5

            elif type == "yz":
                section_span = (
                                       quarter_chord_vector[1] ** 2 +
                                       quarter_chord_vector[2] ** 2
                               ) ** 0.5

            elif type == "xz":
                section_span = (
                                       quarter_chord_vector[0] ** 2 +
                                       quarter_chord_vector[2] ** 2
                               ) ** 0.5

            elif type == "x":
                section_span = quarter_chord_vector[0]

            elif type == "y":
                section_span = quarter_chord_vector[1]

            elif type == "z":
                section_span = quarter_chord_vector[2]

            else:
                raise ValueError("Bad value of 'type'!")

            sectional_spans.append(section_span)

        if _sectional:
            return sectional_spans

        half_span = sum(sectional_spans)

        if include_centerline_distance and len(self.xsecs) > 0:

            half_span_to_XZ_plane = np.inf

            for i in i_range:
                half_span_to_XZ_plane = np.minimum(
                    half_span_to_XZ_plane,
                    np.abs(quarter_chord_locations[i][1])
                )

            half_span = half_span + half_span_to_XZ_plane

        if self.symmetric:
            span = 2 * half_span
        else:
            span = half_span

        return span

    def area(self,
             type: str = "planform",
             include_centerline_distance=False,
             _sectional: bool = False,
             ) -> Union[float, List[float]]:
        """
        Computes the wing area, with options for various ways of measuring this (see `type` argument):

        If the wing is symmetric, both left/right sides are included in order to obtain the full area. In the case
        where the root cross-section is not coincident with the center plane (e.g., XZ plane), this function's
        behavior depends on the `include_centerline_distance` argument.

        Args:

            type: One of the following options, as a string:

                * "planform" (default): First, lofts a quadrilateral mean camber surface between each WingXSec. Then,
                computes the area of each of these sectional surfaces. Then, sums up all the areas and returns it.
                When airplane designers refer to "wing area" (in the absence of any other qualifiers),
                this is typically what they mean.

                * "wetted": Computes the actual surface area of the wing that is in contact with the air. Will
                typically be a little more than double the "planform" area above; intuitively, this is because it
                adds both the "top" and "bottom" surface areas. Accounts for airfoil thickness/shape effects.

                * "xy" or "projected" or "top": Same as "planform", but each sectional surface is projected onto the XY plane
                (i.e., top-down view) before computing the areas. Note that if you try to use this method with a
                vertically-oriented wing, like most vertical stabilizers, you will get an area near zero.

                * "xz" or "side": Same as "planform", but each sectional surface is projected onto the XZ plane before
                computing the areas.

            include_centerline_distance: A boolean flag that tells the function what to do if a wing's root chord is
            not coincident with the centerline plane (i.e., XZ plane).

                * If True, we first figure out which WingXSec is closest to the centerline plane (i.e., XZ plane).
                Then, we imagine that this WingXSec is extruded along the Y axis to the centerline plane (assuming a
                straight extrusion to produce a rectangular mid-camber surface). In doing so, we use the wing
                geometric chord as the extrusion width. We then add the area of this fictitious surface to the area
                calculation.

                * If False, this function will simply ignore this fictitious wing area. This is the default behavior.

            _sectional: A boolean. If False, returns the total area. If True, returns a list of areas for each of the
                `n-1` lofted sections (between the `n` wing cross-sections in wing.xsec).

        """
        # Check inputs
        if include_centerline_distance and _sectional:
            raise ValueError("`include_centerline_distance` and `_sectional` cannot both be True!")

        # Handle overloaded names
        if type == "projected" or type == "top":
            type = "xy"

        elif type == "side":
            type = "xz"

        # Compute sectional areas. Each method must compute the sectional spans and the effective chords at each
        # cross-section to use.
        if type == "planform":
            sectional_spans = self.span(type="yz", _sectional=True)
            xsec_chords = [xsec.chord for xsec in self.xsecs]

        elif type == "wetted":
            sectional_spans = self.span(type="yz", _sectional=True)
            xsec_chords = [
                xsec.chord * xsec.airfoil.perimeter()
                for xsec in self.xsecs
            ]

        elif type == "xy":
            sectional_spans = self.span(type="y", _sectional=True)
            xsec_chords = [xsec.chord for xsec in self.xsecs]

        elif type == "yz":
            raise ValueError("Area of wing projected to the YZ plane is zero.")

            # sectional_spans = self.span(type="yz", _sectional=True)
            # xsec_chords = [xsec.chord for xsec in self.xsecs]

        elif type == "xz":
            sectional_spans = self.span(type="z", _sectional=True)
            xsec_chords = [xsec.chord for xsec in self.xsecs]

        else:
            raise ValueError("Bad value of `type`!")

        sectional_chords = [
            (inner_chord + outer_chord) / 2
            for inner_chord, outer_chord in zip(
                xsec_chords[1:],
                xsec_chords[:-1]
            )
        ]

        sectional_areas = [
            span * chord
            for span, chord in zip(
                sectional_spans,
                sectional_chords
            )
        ]
        if _sectional:
            return sectional_areas

        half_area = sum(sectional_areas)

        if include_centerline_distance and len(self.xsecs) > 0:

            half_span_to_centerline = np.inf

            for i in range(len(self.xsecs)):
                quarter_chord_location = self._compute_xyz_of_WingXSec(
                    i,
                    x_nondim=0.25,
                    z_nondim=0,
                )

                half_span_to_centerline = np.minimum(
                    half_span_to_centerline,
                    np.abs(quarter_chord_location[1])
                )

            half_area = half_area + (
                    half_span_to_centerline * self.mean_geometric_chord()
            )

        if self.symmetric:  # Returns the total area of both the left and right wing halves on mirrored wings.
            area = 2 * half_area
        else:
            area = half_area

        return area

    def aspect_ratio(self,
                     type: str = "geometric",
                     ) -> float:
        """
        Computes the aspect ratio of the wing, with options for various ways of measuring this.

         * geometric: geometric aspect ratio, computed in the typical fashion (b^2 / S).

         * effective: Differs from the geometric aspect ratio only in the case of symmetric wings whose root
         cross-section is not on the centerline. In these cases, it includes the span and area of the fictitious wing
         center when computing aspect ratio.

        Args:
            type: One of the above options, as a string.

        """
        if type == "geometric":
            return self.span() ** 2 / self.area()

        elif type == "effective":
            return (
                    self.span(type="yz", include_centerline_distance=True) ** 2 /
                    self.area(type="planform", include_centerline_distance=True)
            )

        else:
            raise ValueError("Bad value of `type`!")

    def is_entirely_symmetric(self) -> bool:
        # Returns a boolean of whether the wing is totally symmetric (i.e.), every xsec has symmetric control surfaces.
        for xsec in self.xsecs:  # To be symmetric, all
            for surf in xsec.control_surfaces:
                if not (surf.symmetric or surf.deflection == 0):
                    return False

        if not self.symmetric:  # If the wing itself isn't mirrored (e.g., vertical stabilizer), check that it's symmetric
            for xsec in self.xsecs:
                if not xsec.xyz_le[1] == 0:  # Surface has to be right on the centerline
                    return False
                if not xsec.twist == 0:  # Surface has to be untwisted
                    return False
                if not np.allclose(xsec.airfoil.local_camber(), 0):  # Surface has to have a symmetric airfoil.
                    return False

        return True

    def mean_geometric_chord(self) -> float:
        """
        Returns the mean geometric chord of the wing (S/b).
        :return:
        """
        return self.area() / self.span()

    def mean_aerodynamic_chord(self) -> float:
        """
        Computes the length of the mean aerodynamic chord of the wing.
        Uses the generalized methodology described here:
            https://core.ac.uk/download/pdf/79175663.pdf

        Returns: The length of the mean aerodynamic chord.
        """
        sectional_areas = self.area(_sectional=True)
        sectional_MAC_lengths = []

        for inner_xsec, outer_xsec in zip(self.xsecs[:-1], self.xsecs[1:]):

            section_taper_ratio = outer_xsec.chord / inner_xsec.chord
            section_MAC_length = (2 / 3) * inner_xsec.chord * (
                    (1 + section_taper_ratio + section_taper_ratio ** 2) /
                    (1 + section_taper_ratio)
            )

            sectional_MAC_lengths.append(section_MAC_length)

        sectional_MAC_length_area_products = [
            MAC * area
            for MAC, area in zip(
                sectional_MAC_lengths,
                sectional_areas,
            )
        ]

        MAC_length = sum(sectional_MAC_length_area_products) / sum(sectional_areas)

        return MAC_length

    def mean_twist_angle(self) -> float:
        r"""
        Returns the mean twist angle (in degrees) of the wing, weighted by area.
        :return: mean twist angle (in degrees)
        """

        sectional_twists = [
            (inner_xsec.twist + outer_xsec.twist) / 2
            for inner_xsec, outer_xsec in zip(
                self.xsecs[1:],
                self.xsecs[:-1]
            )
        ]
        sectional_areas = self.area(_sectional=True)

        sectional_twist_area_products = [
            twist * area
            for twist, area in zip(
                sectional_twists, sectional_areas
            )
        ]

        mean_twist = sum(sectional_twist_area_products) / sum(sectional_areas)

        return mean_twist

    def mean_sweep_angle(self,
                         x_nondim=0.25
                         ) -> float:
        """
        Returns the mean sweep angle (in degrees) of the wing, relative to the x-axis.
        Positive sweep is backwards, negative sweep is forward.

        This is purely measured from root to tip, with no consideration for the sweep of the individual
        cross-sections in between.

        Args:

            x_nondim: The nondimensional x-coordinate of the cross-section to use for sweep angle computation.

                * If you provide 0, it will use the leading edge of the cross-section.

                * If you provide 0.25, it will use the quarter-chord point of the cross-section.

                * If you provide 1, it will use the trailing edge of the cross-section.

        Returns:

            The mean sweep angle, in degrees.
        """
        root_quarter_chord = self._compute_xyz_of_WingXSec(
            0,
            x_nondim=x_nondim,
            z_nondim=0
        )
        tip_quarter_chord = self._compute_xyz_of_WingXSec(
            -1,
            x_nondim=x_nondim,
            z_nondim=0
        )

        vec = tip_quarter_chord - root_quarter_chord
        vec_norm = vec / np.linalg.norm(vec)

        sin_sweep = vec_norm[0]  # from dot product with x_hat

        sweep_deg = np.arcsind(sin_sweep)

        return sweep_deg

    def mean_dihedral_angle(self,
                            x_nondim=0.25
                            ) -> float:
        """
        Returns the mean dihedral angle (in degrees) of the wing, relative to the XY plane.
        Positive dihedral is bending up, negative dihedral is bending down.

        This is purely measured from root to tip, with no consideration for the dihedral of the individual
        cross-sections in between.

        Args:

            x_nondim: The nondimensional x-coordinate of the cross-section to use for sweep angle computation.

                * If you provide 0, it will use the leading edge of the cross-section.

                * If you provide 0.25, it will use the quarter-chord point of the cross-section.

                * If you provide 1, it will use the trailing edge of the cross-section.

        Returns:

            The mean dihedral angle, in degrees

        """
        root_quarter_chord = self._compute_xyz_of_WingXSec(
            0,
            x_nondim=x_nondim,
            z_nondim=0
        )
        tip_quarter_chord = self._compute_xyz_of_WingXSec(
            -1,
            x_nondim=x_nondim,
            z_nondim=0
        )

        vec = tip_quarter_chord - root_quarter_chord
        vec_norm = vec / np.linalg.norm(vec)

        return np.arctan2d(
            vec_norm[2],
            vec_norm[1],
        )

    def aerodynamic_center(self, chord_fraction: float = 0.25, _sectional=False) -> np.ndarray:
        """
        Computes the location of the aerodynamic center of the wing.
        Uses the generalized methodology described here:
            https://core.ac.uk/downloattttd/pdf/79175663.pdf

        Args: chord_fraction: The position of the aerodynamic center along the MAC, as a fraction of MAC length.

            Typically, this value (denoted `h_0` in the literature) is 0.25 for a subsonic wing. However,
            wing-fuselage interactions can cause a forward shift to a value more like 0.1 or less. Citing Cook,
            Michael V., "Flight Dynamics Principles", 3rd Ed., Sect. 3.5.3 "Controls-fixed static stability". PDF:
            https://www.sciencedirect.com/science/article/pii/B9780080982427000031

        Returns: The (x, y, z) coordinates of the aerodynamic center of the wing.

        """
        sectional_areas = self.area(_sectional=True)
        sectional_ACs = []

        for inner_xsec, outer_xsec in zip(self.xsecs[:-1], self.xsecs[1:]):

            section_taper_ratio = outer_xsec.chord / inner_xsec.chord
            section_MAC_length = (2 / 3) * inner_xsec.chord * (
                    (1 + section_taper_ratio + section_taper_ratio ** 2) /
                    (1 + section_taper_ratio)
            )
            section_MAC_le = (
                    inner_xsec.xyz_le +
                    (outer_xsec.xyz_le - inner_xsec.xyz_le) *
                    (1 + 2 * section_taper_ratio) /
                    (3 + 3 * section_taper_ratio)
            )
            section_AC = section_MAC_le + np.array([  # TODO rotate this vector by the local twist angle
                chord_fraction * section_MAC_length,
                0,
                0
            ])

            sectional_ACs.append(section_AC)

        if _sectional:
            return sectional_ACs

        sectional_AC_area_products = [
            AC * area
            for AC, area in zip(
                sectional_ACs,
                sectional_areas,
            )
        ]

        aerodynamic_center = sum(sectional_AC_area_products) / sum(sectional_areas)

        if self.symmetric:
            aerodynamic_center[1] = 0

        return aerodynamic_center

    def taper_ratio(self) -> float:
        """
        Gives the taper ratio of the Wing. Strictly speaking, only valid for trapezoidal wings.

        Returns:
            Taper ratio of the Wing.

        """
        return self.xsecs[-1].chord / self.xsecs[0].chord

    def volume(self,
               _sectional: bool = False
               ) -> Union[float, List[float]]:
        """
        Computes the volume of the Wing.

        Args:

            _sectional: A boolean. If False, returns the total volume. If True, returns a list of volumes for each of
            the `n-1` lofted sections (between the `n` wing cross-sections in wing.xsec).

        Returns:

            The computed volume.
        """
        xsec_areas = [
            xsec.xsec_area()
            for xsec in self.xsecs
        ]
        separations = self.span(
            type="yz",
            _sectional=True
        )

        sectional_volumes = [
            separation / 3 * (area_a + area_b + (area_a * area_b + 1e-100) ** 0.5)
            for area_a, area_b, separation in zip(
                xsec_areas[1:],
                xsec_areas[:-1],
                separations
            )
        ]

        volume = sum(sectional_volumes)

        if self.symmetric:
            volume *= 2

        if _sectional:
            return sectional_volumes
        else:
            return volume

    def get_control_surface_names(self) -> List[str]:
        """
        Gets the names of all control surfaces on this wing.

        Returns:

            A list of control surface names.

        """
        control_surface_names = []
        for xsec in self.xsecs:
            for control_surface in xsec.control_surfaces:
                control_surface_names.append(control_surface.name)

        return control_surface_names

    def set_control_surface_deflections(self,
                                        control_surface_mappings: Dict[str, float],
                                        ) -> None:
        """
        Sets the deflection of all control surfaces on this wing, based on the provided mapping.

        Args:
            control_surface_mappings: A dictionary mapping control surface names to their deflection angles, in degrees.

                Note: control surface names are set in the asb.ControlSurface constructor.

        Returns:

            None. (in-place)
        """
        for xsec in self.xsecs:
            for control_surface in xsec.control_surfaces:
                if control_surface.name in control_surface_mappings.keys():
                    control_surface.deflection = control_surface_mappings[control_surface.name]

    def control_surface_area(self,
                             by_name: Optional[str] = None,
                             type: Optional[str] = "planform",
                             ) -> float:
        """
        Computes the total area of all control surfaces on this wing, optionally filtered by their name.

        Control surfaces are defined on a section-by-section basis, and are defined in the WingXSec constructor using
        its `control_surfaces` argument.

        Note: If redundant control surfaces are defined (e.g., elevons, as defined by separate ailerons + elevator),
        the area will be duplicated.

        If the wing is symmetric, control surfaces on both left/right sides are included in order to obtain the full area.

        Args:

            by_name: If not None, only control surfaces with this name will be included in the area calculation.

                Note: control surface names are set in the asb.ControlSurface constructor.

            type: One of the following options, as a string:

                * "planform" (default): First, lofts a quadrilateral mean camber surface between each WingXSec. Then,
                computes the area of each of these sectional surfaces. Then, computes what fraction of this area is
                control surface. Then, sums up all the areas and returns it. When airplane designers refer to
                "control surface area" (in the absence of any other qualifiers), this is typically what they mean.

                * "wetted": Computes the actual surface area of the control surface that is in contact with the air.
                Will typically be a little more than double the "planform" area above; intuitively, this is because
                it adds both the "top" and "bottom" surface areas. Accounts for airfoil thickness/shape effects.

                * "xy" or "projected" or "top": Same as "planform", but each sectional surface is projected onto the XY plane
                (i.e., top-down view) before computing the areas. Note that if you try to use this method with a
                vertically-oriented wing, like most vertical stabilizers, you will get an area near zero.

                * "xz" or "side": Same as "planform", but each sectional surface is projected onto the XZ plane before
                computing the areas.

        """
        sectional_areas = self.area(
            type=type,
            include_centerline_distance=False,
            _sectional=True
        )

        control_surface_area = 0.

        for xsec, sect_area in zip(self.xsecs[:-1], sectional_areas):
            for control_surface in xsec.control_surfaces:
                if (by_name is None) or (control_surface.name == by_name):

                    if control_surface.trailing_edge:
                        control_surface_chord_fraction = np.maximum(
                            1 - control_surface.hinge_point,
                            0
                        )
                    else:
                        control_surface_chord_fraction = np.maximum(
                            control_surface.hinge_point,
                            0
                        )

                    control_surface_area += control_surface_chord_fraction * sect_area

        if self.symmetric:
            control_surface_area *= 2

        return control_surface_area

    def mesh_body(self,
                  method="quad",
                  chordwise_resolution: int = 36,
                  chordwise_spacing_function_per_side: Callable[[float, float, float], np.ndarray] = np.cosspace,
                  mesh_surface: bool = True,
                  mesh_tips: bool = True,
                  mesh_trailing_edge: bool = True,
                  mesh_symmetric: bool = True,
                  ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Meshes the outer mold line surface of the wing.

        Uses the `(points, faces)` standard mesh format. For reference on this format, see the documentation in
        `aerosandbox.geometry.mesh_utilities`.

        Order of faces:

            * On the right wing (or, if `Wing.symmetric` is `False`, just the wing itself):

                * If `mesh_surface` is `True`:

                    * First face is nearest the top-side trailing edge of the wing root.

                    * Proceeds chordwise, along the upper surface of the wing from back to front. Upon reaching the
                    leading edge, continues along the lower surface of the wing from front to back.

                    * Then, repeats this process for the next spanwise slice of the wing, and so on.

                * If `mesh_trailing_edge` is `True`:

                    * Continues by meshing the trailing edge of the wing. Meshes the inboard trailing edge first, then
                    proceeds spanwise to the outboard trailing edge.

                * If `mesh_tips` is `True`:

                    * Continues by meshing the wing tips. Meshes the inboard tip first, then meshes the outboard tip.

                    * Within each tip, meshes from the

        Args:

            method: One of the following options, as a string:

                * "tri": Triangular mesh.

                * "quad": Quadrilateral mesh.

            chordwise_resolution: Number of points to use per wing chord, per wing section.

            chordwise_spacing_function_per_side: A function that determines how to space points in the chordwise
            direction along the top and bottom surfaces. Common values would be `np.linspace` or `np.cosspace`,
            but it can be any function with the call signature `f(a, b, n)` that returns a spaced array of `n` points
            between `a` and `b`. [function]

            mesh_surface: If True, includes the actual wing surface in the mesh.

            mesh_tips: If True, includes the wing tips (both on the inboard-most section and on the outboard-most
            section) in the mesh.

            mesh_trailing_edge: If True, includes the wing trailing edge in the mesh, if the trailing-edge thickness
            is nonzero.

            mesh_symmetric: Has no effect if the wing is not symmetric. If the wing is symmetric this determines whether
            the generated mesh is also symmetric, or if if only one side of the wing (right side) is meshed.

        Returns: Standard unstructured mesh format: A tuple of `points` and `faces`, where:

            * `points` is a `n x 3` array of points, where `n` is the number of points in the mesh.

            * `faces` is a `m x 3` array of faces if `method` is "tri", or a `m x 4` array of faces if `method` is "quad".

                * Each row of `faces` is a list of indices into `points`, which specifies a face.

        """

        airfoil_nondim_coordinates = np.array([
            xsec.airfoil
            .repanel(
                n_points_per_side=chordwise_resolution + 1,
                spacing_function_per_side=chordwise_spacing_function_per_side,
            )
            .coordinates
            for xsec in self.xsecs
        ])

        x_nondim = airfoil_nondim_coordinates[:, :, 0].T
        y_nondim = airfoil_nondim_coordinates[:, :, 1].T

        spanwise_strips = []
        for x_n, y_n in zip(x_nondim, y_nondim):
            spanwise_strips.append(
                np.stack(
                    self.mesh_line(
                        x_nondim=x_n,
                        z_nondim=y_n,
                        add_camber=False,
                    ),
                    axis=0
                )
            )

        points = np.concatenate(spanwise_strips, axis=0)

        faces = []

        num_i = (len(self.xsecs) - 1)
        num_j = len(spanwise_strips) - 1

        def index_of(iloc, jloc):
            return iloc + jloc * (num_i + 1)

        def add_face(*indices):
            entry = list(indices)
            if method == "quad":
                faces.append(entry)
            elif method == "tri":
                faces.append([entry[0], entry[1], entry[3]])
                faces.append([entry[1], entry[2], entry[3]])

        if mesh_surface:
            for i in range(num_i):
                for j in range(num_j):
                    add_face(
                        index_of(i, j),
                        index_of(i + 1, j),
                        index_of(i + 1, j + 1),
                        index_of(i, j + 1),
                    )

        if mesh_tips:
            for j in range(num_j // 2):
                add_face(  # Mesh the root face
                    index_of(0, num_j - j),
                    index_of(0, j),
                    index_of(0, j + 1),
                    index_of(0, num_j - j - 1),
                )
                add_face(  # Mesh the tip face
                    index_of(num_i, j),
                    index_of(num_i, j + 1),
                    index_of(num_i, num_j - j - 1),
                    index_of(num_i, num_j - j),
                )
        if mesh_trailing_edge:
            for i in range(num_i):
                add_face(
                    index_of(i + 1, 0),
                    index_of(i + 1, num_j),
                    index_of(i, num_j),
                    index_of(i, 0),
                )

        faces = np.array(faces)

        if mesh_symmetric and self.symmetric:
            flipped_points = np.multiply(
                points,
                np.array([
                    [1, -1, 1]
                ])
            )

            points, faces = mesh_utils.stack_meshes(
                (points, faces),
                (flipped_points, faces)
            )

        return points, faces

    def mesh_thin_surface(self,
                          method="tri",
                          chordwise_resolution: int = 36,
                          chordwise_spacing_function: Callable[[float, float, float], np.ndarray] = np.cosspace,
                          add_camber: bool = True,
                          ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Meshes the mean camber line of the wing as a thin-sheet body.

        Uses the `(points, faces)` standard mesh format. For reference on this format, see the documentation in
        `aerosandbox.geometry.mesh_utilities`.

        Order of faces:
            * On the right wing (or, if `Wing.symmetric` is `False`, just the wing itself):
                * First face is the face nearest the leading edge of the wing root.
                * Proceeds along a chordwise strip to the trailing edge.
                * Then, goes to the subsequent spanwise location and does another chordwise strip, et cetera until
                  we get to the wing tip.
            * On the left wing (applicable only if `Wing.symmetric` is `True`):
                * Same order: Starts at the root leading edge, goes in chordwise strips.

        Order of vertices within each face:
            * On the right wing (or, if `Wing.symmetric` is `False`, just the wing itself):
                * Front-left
                * Back-left
                * Back-right
                * Front-right
            * On the left wing (applicable only if `Wing.symmetric` is `True`):
                * Front-left
                * Back-left
                * Back-right
                * Front-right

        Args:

            method: A string, which determines whether to mesh the fuselage as a series of quadrilaterals or triangles.

                * "quad" meshes the fuselage as a series of quadrilaterals.

                * "tri" meshes the fuselage as a series of triangles.

            chordwise_resolution: Determines the number of chordwise panels to use in the meshing. [int]

            chordwise_spacing_function: Determines how to space the chordwise panels. Can be `np.linspace` or
            `np.cosspace`, or any other function of the call signature `f(a, b, n)` that returns a spaced array of
            `n` points between `a` and `b`. [function]

            add_camber: Controls whether to mesh the thin surface with camber (i.e., mean camber line), or to just
            mesh the flat planform. [bool]

        Returns: Standard unstructured mesh format: A tuple of `points` and `faces`, where:

            * `points` is a `n x 3` array of points, where `n` is the number of points in the mesh.

            * `faces` is a `m x 3` array of faces if `method` is "tri", or a `m x 4` array of faces if `method` is "quad".

                * Each row of `faces` is a list of indices into `points`, which specifies a face.


        """
        x_nondim = chordwise_spacing_function(
            0,
            1,
            chordwise_resolution + 1
        )

        spanwise_strips = []
        for x_n in x_nondim:
            spanwise_strips.append(
                np.stack(
                    self.mesh_line(
                        x_nondim=x_n,
                        z_nondim=0,
                        add_camber=add_camber,
                    ),
                    axis=0
                )
            )

        points = np.concatenate(spanwise_strips)

        faces = []

        num_i = np.length(spanwise_strips[0])  # spanwise
        num_j = np.length(spanwise_strips)  # chordwise

        def index_of(iloc, jloc):
            return iloc + jloc * num_i

        def add_face(*indices):
            entry = list(indices)
            if method == "quad":
                faces.append(entry)
            elif method == "tri":
                faces.append([entry[0], entry[1], entry[3]])
                faces.append([entry[1], entry[2], entry[3]])

        for i in range(num_i - 1):
            for j in range(num_j - 1):
                add_face(  # On right wing:
                    index_of(i, j),  # Front-left
                    index_of(i, j + 1),  # Back-left
                    index_of(i + 1, j + 1),  # Back-right
                    index_of(i + 1, j),  # Front-right
                )

        if self.symmetric:
            index_offset = np.length(points)

            points = np.concatenate([
                points,
                np.multiply(points, np.array([[1, -1, 1]]))
            ])

            def index_of(iloc, jloc):
                return index_offset + iloc + jloc * num_i

            for i in range(num_i - 1):
                for j in range(num_j - 1):
                    add_face(  # On left wing:
                        index_of(i + 1, j),  # Front-left
                        index_of(i + 1, j + 1),  # Back-left
                        index_of(i, j + 1),  # Back-right
                        index_of(i, j),  # Front-right
                    )

        faces = np.array(faces)

        return points, faces

    def mesh_line(self,
                  x_nondim: Union[float, List[float]] = 0.25,
                  z_nondim: Union[float, List[float]] = 0,
                  add_camber: bool = True,
                  ) -> List[np.ndarray]:
        """
        Meshes a line that goes through each of the WingXSec objects in this wing.

        Args:

            x_nondim: The nondimensional (chord-normalized) x-coordinate that the line should go through. Can either
            be a single value used at all cross-sections, or can be an iterable of values to be used at the
            respective cross-sections.

            z_nondim: The nondimensional (chord-normalized) y-coordinate that the line should go through. Here,
            y-coordinate means the "vertical" component (think standard 2D airfoil axes). Can either be a single
            value used at all cross-sections, or can be an iterable of values to be used at the respective cross
            sections.

            add_camber: Controls whether the camber of each cross-section's airfoil should be added to the line or
            not. Essentially modifies `z_nondim` to be `z_nondim + camber`.

        Returns: A list of points, where each point is a 3-element array of the form `[x, y, z]`. Goes from the root
        to the tip. Ignores any wing symmetry (e.g., only gives one side).

        """
        points_on_line: List[np.ndarray] = []

        try:
            if len(x_nondim) != len(self.xsecs):
                raise ValueError(
                    f"If `x_nondim` is an iterable, it should be the same length as `Wing.xsecs` ({len(self.xsecs)})."
                )
        except TypeError:
            pass

        try:
            if len(z_nondim) != len(self.xsecs):
                raise ValueError(
                    f"If `z_nondim` is an iterable, it should be the same length as `Wing.xsecs` ({len(self.xsecs)})."
                )
        except TypeError:
            pass

        for i, xsec in enumerate(self.xsecs):

            try:
                xsec_x_nondim = x_nondim[i]
            except (TypeError, IndexError):
                xsec_x_nondim = x_nondim

            try:
                xsec_z_nondim = z_nondim[i]
            except (TypeError, IndexError):
                xsec_z_nondim = z_nondim

            if add_camber:
                xsec_z_nondim = xsec_z_nondim + xsec.airfoil.local_camber(x_over_c=x_nondim)

            points_on_line.append(
                self._compute_xyz_of_WingXSec(
                    i,
                    x_nondim=xsec_x_nondim,
                    z_nondim=xsec_z_nondim,
                )
            )

        return points_on_line

    def draw(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw()
            **kwargs: Keyword arguments to pass through to Airplane.draw()

        Returns: Same return as Airplane.draw()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(wings=[self]).draw(*args, **kwargs)

    def draw_wireframe(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw_wireframe() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw_wireframe()
            **kwargs: Keyword arguments to pass through to Airplane.draw_wireframe()

        Returns: Same return as Airplane.draw_wireframe()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(wings=[self]).draw_wireframe(*args, **kwargs)

    def draw_three_view(self, *args, **kwargs):
        """
        An alias to the more general Airplane.draw_three_view() method. See there for documentation.

        Args:
            *args: Arguments to pass through to Airplane.draw_three_view()
            **kwargs: Keyword arguments to pass through to Airplane.draw_three_view()

        Returns: Same return as Airplane.draw_three_view()

        """
        from aerosandbox.geometry.airplane import Airplane
        return Airplane(wings=[self]).draw_three_view(*args, **kwargs)

    def subdivide_sections(self,
                           ratio: int,
                           spacing_function: Callable[[float, float, float], np.ndarray] = np.linspace
                           ) -> "Wing":
        """
        Generates a new Wing that subdivides the existing sections of this Wing into several smaller ones. Splits
        each section into N=`ratio` smaller sub-sections by inserting new cross-sections (xsecs) as needed.

        This can allow for finer aerodynamic resolution of sectional properties in certain analyses.

        Args:

            ratio: The number of new sections to split each old section into.

            spacing_function: A function that takes in three arguments: the start, end, and number of points to generate.

                The default is `np.linspace`, which generates a linearly-spaced array of points.

                Other options include `np.cosspace`, which generates a cosine-spaced array of points.

        Returns: A new Wing object with subdivided sections.

        """
        if not (ratio >= 2 and isinstance(ratio, int)):
            raise ValueError("`ratio` must be an integer greater than or equal to 2.")

        new_xsecs = []
        span_fractions_along_section = spacing_function(0, 1, ratio + 1)[:-1]

        for xsec_a, xsec_b in zip(self.xsecs[:-1], self.xsecs[1:]):
            for s in span_fractions_along_section:
                a_weight = 1 - s
                b_weight = s

                if xsec_a.airfoil == xsec_b.airfoil:
                    blended_airfoil = xsec_a.airfoil
                elif a_weight == 1:
                    blended_airfoil = xsec_a.airfoil
                elif b_weight == 1:
                    blended_airfoil = xsec_b.airfoil
                else:
                    blended_airfoil = xsec_a.airfoil.blend_with_another_airfoil(
                        airfoil=xsec_b.airfoil,
                        blend_fraction=b_weight
                    )

                new_xsecs.append(
                    WingXSec(
                        xyz_le=xsec_a.xyz_le * a_weight + xsec_b.xyz_le * b_weight,
                        chord=xsec_a.chord * a_weight + xsec_b.chord * b_weight,
                        twist=xsec_a.twist * a_weight + xsec_b.twist * b_weight,
                        airfoil=blended_airfoil,
                        control_surfaces=xsec_a.control_surfaces,
                        analysis_specific_options=xsec_a.analysis_specific_options,
                    )
                )

        new_xsecs.append(self.xsecs[-1])

        return Wing(
            name=self.name,
            xsecs=new_xsecs,
            symmetric=self.symmetric,
            analysis_specific_options=self.analysis_specific_options
        )

    def _compute_xyz_le_of_WingXSec(self, index: int):
        return self.xsecs[index].xyz_le

    def _compute_xyz_te_of_WingXSec(self, index: int):
        return self._compute_xyz_of_WingXSec(
            index,
            x_nondim=1,
            z_nondim=0,
        )

    def _compute_xyz_of_WingXSec(self,
                                 index,
                                 x_nondim,
                                 z_nondim,
                                 ):
        xg_local, yg_local, zg_local = self._compute_frame_of_WingXSec(index)
        origin = self.xsecs[index].xyz_le
        xsec = self.xsecs[index]
        return origin + (
                x_nondim * xsec.chord * xg_local +
                z_nondim * xsec.chord * zg_local
        )

    def _compute_frame_of_WingXSec(
            self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the local reference frame associated with a particular cross-section (XSec) of this wing.

        Args:

            index: Which cross-section (as indexed in Wing.xsecs) should we get the frame of?

        Returns:

            A tuple of (xg_local, yg_local, zg_local), where each entry refers to the respective (normalized) axis of
            the local reference frame of the WingXSec. Given in geometry axes.

        """

        def project_to_YZ_plane_and_normalize(vector):
            YZ_magnitude = (vector[1] ** 2 + vector[2] ** 2) ** 0.5
            return np.array([0, vector[1], vector[2]]) / YZ_magnitude

        ### Compute the untwisted reference frame
        xg_local = np.array([1, 0, 0])
        if index == 0:
            yg_local = project_to_YZ_plane_and_normalize(
                self.xsecs[1].xyz_le - self.xsecs[0].xyz_le
            )
            z_scale = 1
        elif index == len(self.xsecs) - 1 or index == -1:
            yg_local = project_to_YZ_plane_and_normalize(
                self.xsecs[-1].xyz_le - self.xsecs[-2].xyz_le
            )
            z_scale = 1
        else:
            vector_before = project_to_YZ_plane_and_normalize(
                self.xsecs[index].xyz_le - self.xsecs[index - 1].xyz_le
            )
            vector_after = project_to_YZ_plane_and_normalize(
                self.xsecs[index + 1].xyz_le - self.xsecs[index].xyz_le
            )
            span_vector = (vector_before + vector_after) / 2
            yg_local = span_vector / np.linalg.norm(span_vector)
            cos_vectors = np.linalg.inner(vector_before, vector_after)
            z_scale = np.sqrt(2 / (cos_vectors + 1))

        zg_local = np.cross(xg_local, yg_local) * z_scale

        ### Twist the reference frame by the WingXSec twist angle
        rot = np.rotation_matrix_3D(
            self.xsecs[index].twist * pi / 180,
            yg_local
        )
        xg_local = rot @ xg_local
        zg_local = rot @ zg_local

        return xg_local, yg_local, zg_local

    def _compute_frame_of_section(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes the local reference frame associated with a particular section. (Note that sections and cross
        sections are different! cross-sections, or xsecs, are the vertices, and sections are the parts in between. In
        other words, a wing with N cross-sections (xsecs) will always have N-1 sections.

        Args:

            index: Which section should we get the frame of? If given `i`, this retrieves the frame of the section
            between xsecs `i` and `i+1`.

        Returns:

            A tuple of (xg_local, yg_local, zg_local), where each entry refers to the respective (normalized) axis
            of the local reference frame of the section. Given in geometry axes.

        """
        in_front = self._compute_xyz_le_of_WingXSec(index)
        in_back = self._compute_xyz_te_of_WingXSec(index)
        out_front = self._compute_xyz_le_of_WingXSec(index + 1)
        out_back = self._compute_xyz_te_of_WingXSec(index + 1)

        diag1 = out_back - in_front
        diag2 = out_front - in_back
        cross = np.cross(diag1, diag2)

        zg_local = cross / np.linalg.norm(cross)

        quarter_chord_vector = (
                                       0.75 * out_front + 0.25 * out_back
                               ) - (
                                       0.75 * in_front + 0.25 * in_back
                               )
        quarter_chord_vector[0] = 0

        yg_local = quarter_chord_vector / np.linalg.norm(quarter_chord_vector)

        xg_local = np.cross(yg_local, zg_local)

        return xg_local, yg_local, zg_local


class WingXSec(AeroSandboxObject):
    """
    Definition for a wing cross-section ("X-section").
    """

    def __init__(self,
                 xyz_le: Union[np.ndarray, List] = None,
                 chord: float = 1.,
                 twist: float = 0.,
                 airfoil: Airfoil = None,
                 control_surfaces: Optional[List['ControlSurface']] = None,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 **deprecated_kwargs,
                 ):
        """
        Defines a new wing cross-section.

        Args:

            xyz_le: An array-like that represents the xyz-coordinates of the leading edge of the cross-section, in
            geometry axes.

            chord: Chord of the wing at this cross-section.

            twist: Twist angle, in degrees, as defined about the leading edge.

                The twist axis is computed with the following procedure:

                    * The quarter-chord point of this WingXSec and the following one are identified.

                    * A line is drawn connecting them, and it is normalized to a unit direction vector.

                    * That direction vector is projected onto the geometry Y-Z plane.

                    * That direction vector is now the twist axis.

            airfoil: Airfoil associated with this cross-section. [aerosandbox.Airfoil]

            control_surfaces: A list of control surfaces in the form of ControlSurface objects.

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }

            Note: Control surface definition through WingXSec properties (control_surface_is_symmetric, control_surface_hinge_point, control_surface_deflection)
            is deprecated. Control surfaces should be handled according to the following protocol:
            
                1. If control_surfaces is an empty list (default, user does not specify any control surfaces), use deprecated WingXSec control surface definition properties.
                This will result in 1 control surface at this xsec.
                
                Usage example:

                >>> xsecs = asb.WingXSec(
                >>>     chord = 2
                >>> )

                2. If control_surfaces is a list of ControlSurface instances, use ControlSurface properties to define control surfaces. This will result in as many
                control surfaces at this xsec as there are entries in the control_surfaces list (an arbitrary number >= 1).

                Usage example:

                >>>xsecs = asb.WingXSec(
                >>>    chord = 2,
                >>>    control_surfaces = [
                >>>        ControlSurface(
                >>>            trailing_edge = False
                >>>        )
                >>>    ]
                >>>)

                3. If control_surfaces is None, override deprecated control surface definition properties and do not define a control surface at this xsec. This will
                result in 0 control surfaces at this xsec.

                Usage example:

                >>>xsecs = asb.WingXSec(
                >>>    chord = 2,
                >>>    control_surfaces = None
                >>>)
            
            See avl.py for example of control_surface handling using this protocol.
        """
        ### Set defaults
        if xyz_le is None:
            xyz_le = np.array([0., 0., 0.])
        if airfoil is None:
            import warnings
            warnings.warn(
                "An airfoil is not specified for WingXSec. Defaulting to NACA 0012.",
                stacklevel=2
            )
            airfoil = Airfoil("naca0012")
        if control_surfaces is None:
            control_surfaces = []
        if analysis_specific_options is None:
            analysis_specific_options = {}

        self.xyz_le = np.array(xyz_le)
        self.chord = chord
        self.twist = twist
        self.airfoil = airfoil
        self.control_surfaces = control_surfaces
        self.analysis_specific_options = analysis_specific_options

        ### Handle deprecated arguments
        if 'twist_angle' in deprecated_kwargs.keys():
            import warnings
            warnings.warn(
                "DEPRECATED: 'twist_angle' has been renamed 'twist', and will break in future versions.",
                stacklevel=2
            )
            self.twist = deprecated_kwargs['twist_angle']
        if (
                'control_surface_is_symmetric' in locals() or
                'control_surface_hinge_point' in locals() or
                'control_surface_deflection' in locals()
        ):
            import warnings
            warnings.warn(
                "DEPRECATED: Define control surfaces using the `control_surfaces` parameter, which takes in a list of asb.ControlSurface objects.",
                stacklevel=2
            )
            if 'control_surface_is_symmetric' not in locals():
                control_surface_is_symmetric = True
            if 'control_surface_hinge_point' not in locals():
                control_surface_hinge_point = 0.75
            if 'control_surface_deflection' not in locals():
                control_surface_deflection = 0

            self.control_surfaces.append(
                ControlSurface(
                    hinge_point=control_surface_hinge_point,
                    symmetric=control_surface_is_symmetric,
                    deflection=control_surface_deflection,
                )
            )

    def __repr__(self) -> str:
        return f"WingXSec (Airfoil: {self.airfoil.name}, chord: {self.chord}, twist: {self.twist})"

    def translate(self,
                  xyz: Union[np.ndarray, List]
                  ) -> "WingXSec":
        """
        Returns a copy of this WingXSec that has been translated by `xyz`.

        Args:
            xyz: The amount to translate the WingXSec. Given as a 3-element NumPy vector.

        Returns: A new WingXSec object.

        """
        new_xsec = copy.copy(self)
        new_xsec.xyz_le = new_xsec.xyz_le + np.array(xyz)
        return new_xsec

    def xsec_area(self):
        """
        Computes the WingXSec's cross-sectional (xsec) area.

        Returns: The (dimensional) cross-sectional area of the WingXSec.
        """
        return self.airfoil.area() * self.chord ** 2


class ControlSurface(AeroSandboxObject):
    """
    Definition for a control surface, which is attached to a particular WingXSec via WingXSec's `control_surfaces=[]` parameter.
    """

    def __init__(self,
                 name: str = "Untitled",
                 symmetric: bool = True,
                 deflection: float = 0.0,
                 hinge_point: float = 0.75,
                 trailing_edge: bool = True,
                 analysis_specific_options: Optional[Dict[type, Dict[str, Any]]] = None,
                 ):
        """
        Define a new control surface.

        Args:

            name: Name of the control surface [optional]. It can help when debugging to give each control surface a
            sensible name.

            symmetric: Is the control surface symmetric? If False, control surface is anti-symmetric. (e.g.,
            True for flaps, False for ailerons.)

            hinge_point: The location of the control surface hinge, as a fraction of chord. A float in the range of 0 to 1.

            deflection: Control deflection, in degrees. Downwards-positive.

            trailing_edge: Is the control surface on the trailing edge? If False, control surface is on the leading
            edge. (e.g., True for flaps, False for slats.). Support is experimental for leading-edge control
            surfaces, be aware that not all modules may treat this correctly.

            analysis_specific_options: Analysis-specific options are additional constants or modeling assumptions
            that should be passed on to specific analyses and associated with this specific geometry object.

                This should be a dictionary where:

                    * Keys are specific analysis types (typically a subclass of asb.ExplicitAnalysis or
                    asb.ImplicitAnalysis), but if you decide to write your own analysis and want to make this key
                    something else (like a string), that's totally fine - it's just a unique identifier for the
                    specific analysis you're running.

                    * Values are a dictionary of key:value pairs, where:

                        * Keys are strings.

                        * Values are some value you want to assign.

                This is more easily demonstrated / understood with an example:

                >>> analysis_specific_options = {
                >>>     asb.AeroBuildup: dict(
                >>>         include_wave_drag=True,
                >>>     )
                >>> }

        """
        ### Set defaults
        if analysis_specific_options is None:
            analysis_specific_options = {}

        self.name = name
        self.symmetric = symmetric
        self.deflection = deflection
        self.hinge_point = hinge_point
        self.trailing_edge = trailing_edge
        self.analysis_specific_options = analysis_specific_options

    def __repr__(self) -> str:
        keys = [
            "name",
            "symmetric",
            "deflection",
            "hinge_point",
        ]
        if not self.trailing_edge:
            keys += ["trailing_edge"]

        info = ", ".join([
            f"{k}={self.__dict__[k]}"
            for k in keys
        ])

        return f"ControlSurface ({info})"


if __name__ == '__main__':
    wing = Wing(
        xsecs=[
            WingXSec(
                xyz_le=[0, 0, 0],
                chord=1,
                airfoil=Airfoil("naca4412"),
                twist=0,
                control_surfaces=[
                    ControlSurface(
                        name="Elevator",
                        trailing_edge=True,
                        hinge_point=0.75,
                        deflection=5
                    )
                ]
            ),
            WingXSec(
                xyz_le=[0.5, 1, 0],
                chord=0.5,
                airfoil=Airfoil("naca4412"),
                twist=0,
            ),
            WingXSec(
                xyz_le=[0.7, 1, 0.3],
                chord=0.3,
                airfoil=Airfoil("naca0012"),
                twist=0,
            )
        ]
    ).translate([1, 0, 0])
    # wing.subdivide_sections(5).draw()



import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
from typing import Tuple, Dict, Union, Callable, List
from scipy import interpolate


def plot_color_by_value(
        x: np.ndarray,
        y: np.ndarray,
        *args,
        c: np.ndarray,
        cmap='turbo',
        colorbar: bool = False,
        colorbar_label: str = None,
        clim: Tuple[float, float] = None,
        **kwargs
):
    """
    Uses same syntax as matplotlib.pyplot.plot, except that `c` is now an array-like that maps to a specific color
    pulled from `cmap`. Makes lines that are multicolored based on this `c` value.

    Args:

        x: Array of x-points.

        y: Array of y-points.

        *args: Args that will be passed into matplotlib.pyplot.plot().
            Example: ".-" for a dotted line.

        c: Array of values that will map to colors. Must be the same length as x and y.

        cmap: The colormap to use.

        colorbar: Whether or not to display the colormap. [bool]

        colorbar_label: The label to add to the colorbar. Only applies if the colorbar is created. [str]

        clim: A tuple of (min, max) that assigns bounds to the colormap. Computed from the range of `c` if not given.

        **kwargs: Kwargs that will be passed into matplotlib.pyplot.plot()


    Returns:

    """
    cmap = mpl.colormaps.get_cmap(cmap)

    x = np.array(x)
    y = np.array(y)
    c = np.array(c)

    cmin = c.min()
    cmax = c.max()

    if clim is None:
        clim = (cmin, cmax)

    norm = plt.Normalize(vmin=clim[0], vmax=clim[1], clip=False)

    label = kwargs.pop("label", None)

    lines = []

    for i, (
            x1, x2,
            y1, y2,
            c1, c2,
    ) in enumerate(zip(
        x[:-1], x[1:],
        y[:-1], y[1:],
        c[:-1], c[1:],
    )):
        line = plt.plot(
            [x1, x2],
            [y1, y2],
            *args,
            color=cmap(
                norm(
                    (c1 + c2) / 2
                ) if cmin != cmax else 0.5
            ),
            **kwargs
        )
        lines += line

    if label is not None:
        line = plt.plot(
            [None],
            [None],
            *args,
            color=cmap(0.5),
            label=label,
            **kwargs
        )
        lines += line
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    if colorbar:
        if colorbar_label is None:
            cbar = plt.colorbar(sm, ax=plt.gca())
        else:
            cbar = plt.colorbar(sm, ax=plt.gca(), label=colorbar_label)
    else:
        cbar = None
    return lines, sm, cbar

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()
    x = np.linspace(-1, 1, 500)
    y = np.sin(10 * x)
    c = np.sin((10 * x) ** 2)
    plot_color_by_value(
        x, y,
        c=c,
        clim=(-1, 1),
        colorbar=True, colorbar_label="Colorbar Label"
    )
    p.show_plot(
        "Title",
        "X Axis",
        "Y Axis",
    )


import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.tools import units as u
from typing import Tuple


### See also:
# https://www.risingup.com/fars/info/23-index.shtml

def limit_load_factors(
        design_mass_TOGW: float,
        category: str = "normal",
) -> Tuple[float, float]:
    """
    Computes the required limit load factors for FAR Part 23 certification.

    From FAR Part 23: "Airworthiness Standards: Normal, Utility, Acrobatic, and Commuter Category Airplanes"
    Section 23.337: "Limit maneuvering load factors"

    Args:

        design_mass_TOGW: The design takeoff gross weight of the aircraft [kg].

        category: The category of the aircraft. Valid values are:

            - "normal"
            - "utility"
            - "acrobatic"
            - "commuter"

    Returns:
        A tuple with (positive load factor, negative load factor). These are the maximum positive and negative limit load
        factors that the aircraft should withstand for Part 23 certification.

    """
    ### Compute positive load factor
    if category == "normal" or category == "commuter":
        positive_load_factor = np.softmin(
            2.1 + (24000 / (design_mass_TOGW / u.lbm + 10000)),
            3.8,
            softness=0.01
        )
    elif category == "utility":
        positive_load_factor = 4.4

    elif category == "acrobatic":
        positive_load_factor = 6.0

    else:
        raise ValueError("Bad value of `category`. Valid values are 'normal', 'utility', 'acrobatic', and 'commuter'.")

    ### Compute negative load factor
    if category == "normal" or category == "commuter" or category == "utility":
        negative_load_factor = -0.4 * positive_load_factor
    elif category == "acrobatic":
        negative_load_factor = -0.5 * positive_load_factor
    else:
        raise ValueError()

    return positive_load_factor, negative_load_factor



import aerosandbox.numpy as np
from scipy.special import comb
import re
from typing import Union
import os
from typing import List, Optional, Dict

_default_n_points_per_side = 200


def get_NACA_coordinates(
        name: str = None,
        n_points_per_side: int = _default_n_points_per_side,
        max_camber: float = None,
        camber_loc: float = None,
        thickness: float = None,
) -> np.ndarray:
    """
    Returns the coordinates of a 4-series NACA airfoil.

    Can EITHER specify `name`, or all three of `max_camber`, `camber_loc`, and `thickness` - not both.

    Args:

        Either:

            * name: Name of the NACA airfoil, as a string (e.g., "naca2412")

        Or:

            * All three of:

                max_camber: Maximum camber of the airfoil, as a fraction of chord (e.g., 0.02)

                camber_loc: The location of maximum camber, as a fraction of chord (e.g., 0.40)

                thickness: The maximum thickness of the airfoil, as a fraction of chord (e.g., 0.12)

        n_points_per_side: Number of points per side of the airfoil (top/bottom).

    Returns: The coordinates of the airfoil as a Nx2 ndarray [x, y]

    """
    ### Validate inputs
    name_specified = name is not None
    params_specified = [
        (max_camber is not None),
        (camber_loc is not None),
        (thickness is not None)
    ]

    if name_specified:
        if any(params_specified):
            raise ValueError(
                "Cannot specify both `name` and (`max_camber`, `camber_loc`, `thickness`) parameters - must be one or the other.")

        name = name.lower().strip()

        if not "naca" in name:
            raise ValueError("Not a NACA airfoil - name must start with 'naca'!")

        nacanumber = name.split("naca")[1]
        if not nacanumber.isdigit():
            raise ValueError("Couldn't parse the number of the NACA airfoil!")

        if not len(nacanumber) == 4:
            raise NotImplementedError("Only 4-digit NACA airfoils are currently supported!")

        # Parse
        max_camber = int(nacanumber[0]) * 0.01
        camber_loc = int(nacanumber[1]) * 0.1
        thickness = int(nacanumber[2:]) * 0.01

    else:
        if not all(params_specified):
            raise ValueError(
                "Must specify either `name` or all three (`max_camber`, `camber_loc`, `thickness`) parameters.")

    # Referencing https://en.wikipedia.org/wiki/NACA_airfoil#Equation_for_a_cambered_4-digit_NACA_airfoil
    # from here on out

    # Make uncambered coordinates
    x_t = np.cosspace(0, 1, n_points_per_side)  # Generate some cosine-spaced points
    y_t = 5 * thickness * (
            + 0.2969 * x_t ** 0.5
            - 0.1260 * x_t
            - 0.3516 * x_t ** 2
            + 0.2843 * x_t ** 3
            - 0.1015 * x_t ** 4  # 0.1015 is original, #0.1036 for sharp TE
    )

    if camber_loc == 0:
        camber_loc = 0.5  # prevents divide by zero errors for things like naca0012's.

    # Get camber
    y_c = np.where(
        x_t <= camber_loc,
        max_camber / camber_loc ** 2 * (2 * camber_loc * x_t - x_t ** 2),
        max_camber / (1 - camber_loc) ** 2 * ((1 - 2 * camber_loc) + 2 * camber_loc * x_t - x_t ** 2)
    )

    # Get camber slope
    dycdx = np.where(
        x_t <= camber_loc,
        2 * max_camber / camber_loc ** 2 * (camber_loc - x_t),
        2 * max_camber / (1 - camber_loc) ** 2 * (camber_loc - x_t)
    )
    theta = np.arctan(dycdx)

    # Combine everything
    x_U = x_t - y_t * np.sin(theta)
    x_L = x_t + y_t * np.sin(theta)
    y_U = y_c + y_t * np.cos(theta)
    y_L = y_c - y_t * np.cos(theta)

    # Flip upper surface so it's back to front
    x_U, y_U = x_U[::-1], y_U[::-1]

    # Trim 1 point from lower surface so there's no overlap
    x_L, y_L = x_L[1:], y_L[1:]

    x = np.concatenate((x_U, x_L))
    y = np.concatenate((y_U, y_L))

    return np.stack((x, y), axis=1)


def get_kulfan_coordinates(
        lower_weights: np.ndarray = -0.2 * np.ones(8),
        upper_weights: np.ndarray = 0.2 * np.ones(8),
        leading_edge_weight: float = 0.,
        TE_thickness: float = 0.,
        n_points_per_side: int = _default_n_points_per_side,
        N1: float = 0.5,
        N2: float = 1.0,
        **deprecated_kwargs
) -> np.ndarray:
    """
    Given a set of Kulfan parameters, computes the coordinates of the resulting airfoil.

    This function is the inverse of `get_kulfan_parameters()`.

    Kulfan parameters are a highly-efficient and flexible way to parameterize the shape of an airfoil. The particular
    flavor of Kulfan parameterization used in AeroSandbox is the "CST with LEM" method, which is described in various
    papers linked below. In total, the Kulfan parameterization consists of:

    * A vector of weights corresponding to the lower surface of the airfoil
    * A vector of weights corresponding to the upper surface of the airfoil
    * A scalar weight corresponding to the strength of a leading-edge camber mode shape of the airfoil (optional)
    * The trailing-edge (TE) thickness of the airfoil (optional)

    These Kulfan parameters are also referred to as CST (Class/Shape Transformation) parameters.

    References on Kulfan (CST) airfoils:

    * Kulfan, Brenda "Universal Parametric Geometry Representation Method" (2008). AIAA Journal of Aircraft.
        Describes the basic Kulfan (CST) airfoil parameterization.
        Mirrors:
            * https://arc.aiaa.org/doi/10.2514/1.29958
            * https://www.brendakulfan.com/_files/ugd/169bff_6738e0f8d9074610942c53dfaea8e30c.pdf
            * https://www.researchgate.net/publication/245430684_Universal_Parametric_Geometry_Representation_Method

    * Kulfan, Brenda "Modification of CST Airfoil Representation Methodology" (2020). Unpublished note:
        Describes the optional "Leading-Edge Modification" (LEM) addition to the Kulfan (CST) airfoil parameterization.
        Mirrors:
            * https://www.brendakulfan.com/_files/ugd/169bff_16a868ad06af4fea946d299c6028fb13.pdf
            * https://www.researchgate.net/publication/343615711_Modification_of_CST_Airfoil_Representation_Methodology

    * Masters, D.A. "Geometric Comparison of Aerofoil Shape Parameterization Methods" (2017). AIAA Journal.
        Compares the Kulfan (CST) airfoil parameterization to other airfoil parameterizations. Also has further notes
        on the LEM addition.
        Mirrors:
            * https://arc.aiaa.org/doi/10.2514/1.J054943
            * https://research-information.bris.ac.uk/ws/portalfiles/portal/91793513/SP_Journal_RED.pdf

    Notes on N1, N2 (shape factor) combinations:
        * 0.5, 1: Conventional airfoil
        * 0.5, 0.5: Elliptic airfoil
        * 1, 1: Biconvex airfoil
        * 0.75, 0.75: Sears-Haack body (radius distribution)
        * 0.75, 0.25: Low-drag projectile
        * 1, 0.001: Cone or wedge airfoil
        * 0.001, 0.001: Rectangle, circular duct, or circular rod.

    To make a Kulfan (CST) airfoil, use the following syntax:

    >>> import aerosandbox as asb
    >>> asb.Airfoil("My Airfoil Name", coordinates=asb.get_kulfan_coordinates(*args))

    Args:

        lower_weights (iterable): The Kulfan weights to use for the lower surface.

        upper_weights (iterable): The Kulfan weights to use for the upper surface.

        TE_thickness (float): The trailing-edge thickness to add, in terms of y/c.

        n_points_per_side (int): The number of points to discretize with, when generating the coordinates.

        N1 (float): The shape factor corresponding to the leading edge of the airfoil. See above for examples.

        N2 (float): The shape factor corresponding to the trailing edge of the airfoil. See above for examples.

    Returns:
        np.ndarray: The coordinates of the airfoil as a Nx2 array.
    """
    if len(deprecated_kwargs) > 0:
        import warnings
        warnings.warn(
            "The following arguments are deprecated and will be removed in a future version:\n"
            f"{deprecated_kwargs}",
            DeprecationWarning
        )

        if deprecated_kwargs.get("enforce_continuous_LE_radius", False):
            lower_weights[0] = -1 * upper_weights[0]

    x = np.cosspace(0, 1, n_points_per_side)  # Generate some cosine-spaced points

    # Class function
    C = (x) ** N1 * (1 - x) ** N2

    def shape_function(w):
        # Shape function (Bernstein polynomials)
        N = np.length(w) - 1  # Order of Bernstein polynomials

        K = comb(N, np.arange(N + 1))  # Bernstein polynomial coefficients

        dims = (np.length(w), np.length(x))

        def wide(vector):
            return np.tile(np.reshape(vector, (1, dims[1])), (dims[0], 1))

        def tall(vector):
            return np.tile(np.reshape(vector, (dims[0], 1)), (1, dims[1]))

        S_matrix = (
                tall(K) * wide(x) ** tall(np.arange(N + 1)) *
                wide(1 - x) ** tall(N - np.arange(N + 1))
        )  # Bernstein polynomial coefficients * weight matrix
        S_x = np.sum(tall(w) * S_matrix, axis=0)

        # Calculate y output
        y = C * S_x
        return y

    y_lower = shape_function(lower_weights)
    y_upper = shape_function(upper_weights)

    # Add trailing-edge (TE) thickness
    y_lower -= x * TE_thickness / 2
    y_upper += x * TE_thickness / 2

    # Add Kulfan's leading-edge-modification (LEM)
    y_lower += leading_edge_weight * (x) * (1 - x) ** (np.length(lower_weights) + 0.5)
    y_upper += leading_edge_weight * (x) * (1 - x) ** (np.length(upper_weights) + 0.5)

    x = np.concatenate((x[::-1], x[1:]))
    y = np.concatenate((y_upper[::-1], y_lower[1:]))
    coordinates = np.stack((x, y), axis=1)

    return coordinates


def get_kulfan_parameters(
        coordinates: np.ndarray,
        n_weights_per_side: int = 8,
        N1: float = 0.5,
        N2: float = 1.0,
        n_points_per_side: int = _default_n_points_per_side,
        normalize_coordinates: bool = True,
        use_leading_edge_modification: bool = True,
        method: str = "least_squares",
) -> Dict[str, Union[np.ndarray, float]]:
    """
    Given a set of airfoil coordinates, reconstructs the Kulfan parameters that would recreate that airfoil. Uses a
    curve fitting (optimization) process.

    This function is the inverse of `get_kulfan_coordinates()`.

    Kulfan parameters are a highly-efficient and flexible way to parameterize the shape of an airfoil. The particular
    flavor of Kulfan parameterization used in AeroSandbox is the "CST with LEM" method, which is described in various
    papers linked below. In total, the Kulfan parameterization consists of:

    * A vector of weights corresponding to the lower surface of the airfoil
    * A vector of weights corresponding to the upper surface of the airfoil
    * A scalar weight corresponding to the strength of a leading-edge camber mode shape of the airfoil (optional)
    * The trailing-edge (TE) thickness of the airfoil (optional)

    These Kulfan parameters are also referred to as CST (Class/Shape Transformation) parameters.

    References on Kulfan (CST) airfoils:

    * Kulfan, Brenda "Universal Parametric Geometry Representation Method" (2008). AIAA Journal of Aircraft.
        Describes the basic Kulfan (CST) airfoil parameterization.
        Mirrors:
            * https://arc.aiaa.org/doi/10.2514/1.29958
            * https://www.brendakulfan.com/_files/ugd/169bff_6738e0f8d9074610942c53dfaea8e30c.pdf
            * https://www.researchgate.net/publication/245430684_Universal_Parametric_Geometry_Representation_Method

    * Kulfan, Brenda "Modification of CST Airfoil Representation Methodology" (2020). Unpublished note:
        Describes the optional "Leading-Edge Modification" (LEM) addition to the Kulfan (CST) airfoil parameterization.
        Mirrors:
            * https://www.brendakulfan.com/_files/ugd/169bff_16a868ad06af4fea946d299c6028fb13.pdf
            * https://www.researchgate.net/publication/343615711_Modification_of_CST_Airfoil_Representation_Methodology

    * Masters, D.A. "Geometric Comparison of Aerofoil Shape Parameterization Methods" (2017). AIAA Journal.
        Compares the Kulfan (CST) airfoil parameterization to other airfoil parameterizations. Also has further notes
        on the LEM addition.
        Mirrors:
            * https://arc.aiaa.org/doi/10.2514/1.J054943
            * https://research-information.bris.ac.uk/ws/portalfiles/portal/91793513/SP_Journal_RED.pdf

    Notes on N1, N2 (shape factor) combinations:
        * 0.5, 1: Conventional airfoil
        * 0.5, 0.5: Elliptic airfoil
        * 1, 1: Biconvex airfoil
        * 0.75, 0.75: Sears-Haack body (radius distribution)
        * 0.75, 0.25: Low-drag projectile
        * 1, 0.001: Cone or wedge airfoil
        * 0.001, 0.001: Rectangle, circular duct, or circular rod.

    The following demonstrates the reversibility of this function:

    >>> import aerosandbox as asb
    >>> from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters
    >>>
    >>> af = asb.Airfoil("dae11")  # A conventional airfoil
    >>> params = get_kulfan_parameters(
    >>>     coordinates=af.coordinates,
    >>> )
    >>> af_reconstructed = asb.Airfoil(
    >>>     name="Reconstructed Airfoil",
    >>>     coordinates=get_kulfan_coordinates(
    >>>         **params
    >>>     )

    Args:

        coordinates (np.ndarray): The coordinates of the airfoil as a Nx2 array.

        n_weights_per_side (int): The number of Kulfan weights to use per side of the airfoil.

        N1 (float): The shape factor corresponding to the leading edge of the airfoil. See above for examples.

        N2 (float): The shape factor corresponding to the trailing edge of the airfoil. See above for examples.

        n_points_per_side (int): The number of points to discretize with, when formulating the curve-fitting
            optimization problem.

    Returns:
        A dictionary containing the Kulfan parameters. The keys are:
            * "lower_weights" (np.ndarray): The weights corresponding to the lower surface of the airfoil.
            * "upper_weights" (np.ndarray): The weights corresponding to the upper surface of the airfoil.
            * "TE_thickness" (float): The trailing-edge thickness of the airfoil.
            * "leading_edge_weight" (float): The strength of the leading-edge camber mode shape of the airfoil.

        These can be passed directly into `get_kulfan_coordinates()` to reconstruct the airfoil.
    """
    from aerosandbox.geometry.airfoil import Airfoil

    if method == "opti":

        target_airfoil = Airfoil(
            name="Target Airfoil",
            coordinates=coordinates
        ).repanel(
            n_points_per_side=n_points_per_side
        )

        if normalize_coordinates:
            target_airfoil = target_airfoil.normalize()

        x = np.cosspace(0, 1, n_points_per_side)
        target_thickness = target_airfoil.local_thickness(x_over_c=x)
        target_camber = target_airfoil.local_camber(x_over_c=x)

        target_y_upper = target_camber + target_thickness / 2
        target_y_lower = target_camber - target_thickness / 2

        # Class function
        C = (x) ** N1 * (1 - x) ** N2

        def shape_function(w):
            # Shape function (Bernstein polynomials)
            N = np.length(w) - 1  # Order of Bernstein polynomials

            K = comb(N, np.arange(N + 1))  # Bernstein polynomial coefficients

            dims = (np.length(w), np.length(x))

            def wide(vector):
                return np.tile(np.reshape(vector, (1, dims[1])), (dims[0], 1))

            def tall(vector):
                return np.tile(np.reshape(vector, (dims[0], 1)), (1, dims[1]))

            S_matrix = (
                    tall(K) * wide(x) ** tall(np.arange(N + 1)) *
                    wide(1 - x) ** tall(N - np.arange(N + 1))
            )  # Bernstein polynomial coefficients * weight matrix
            S_x = np.sum(tall(w) * S_matrix, axis=0)

            # Calculate y output
            y = C * S_x
            return y

        opti = asb.Opti()
        lower_weights = opti.variable(init_guess=0, n_vars=n_weights_per_side)
        upper_weights = opti.variable(init_guess=0, n_vars=n_weights_per_side)
        TE_thickness = opti.variable(init_guess=0, lower_bound=0)
        if use_leading_edge_modification:
            leading_edge_weight = opti.variable(init_guess=0)
        else:
            leading_edge_weight = 0

        y_lower = shape_function(lower_weights)
        y_upper = shape_function(upper_weights)

        # Add trailing-edge (TE) thickness
        y_lower -= x * TE_thickness / 2
        y_upper += x * TE_thickness / 2

        # Add Kulfan's leading-edge-modification (LEM)
        y_lower += leading_edge_weight * (x) * (1 - x) ** (np.length(lower_weights) + 0.5)
        y_upper += leading_edge_weight * (x) * (1 - x) ** (np.length(upper_weights) + 0.5)

        opti.minimize(
            np.sum((y_lower - target_y_lower) ** 2) +
            np.sum((y_upper - target_y_upper) ** 2)
        )

        sol = opti.solve(
            verbose=False
        )

        return {
            "lower_weights"      : sol(lower_weights),
            "upper_weights"      : sol(upper_weights),
            "TE_thickness"       : sol(TE_thickness),
            "leading_edge_weight": sol(leading_edge_weight),
        }

    elif method == "least_squares":

        """
        
        The goal here is to set up this fitting problem as a least-squares problem (likely an overconstrained one, 
        but keeping it general for now. This will then be solved with np.linalg.lstsq(A, b), where A will (likely) 
        not be square.
        
        The columns of the A matrix will correspond to our unknowns, which are going to be a 1D vector `x` packed in as:
            * upper_weights from 0 to n_weights_per_side - 1
            * lower_weights from 0 to n_weights_per_side - 1
            * leading_edge_weight
            * trailing_edge_thickness
            
        See `get_kulfan_coordinates()` for more details on the meaning of these variables.
        
        The rows of the A matrix will correspond to each row of the given airfoil coordinates (i.e., a single vertex 
        on the airfoil). The idea here is to express each vertex as a linear combination of the unknowns, and then
        solve for the unknowns that minimize the error between the given airfoil coordinates and the reconstructed
        airfoil coordinates.
        
        """

        if normalize_coordinates:
            coordinates = Airfoil(
                name="Target Airfoil",
                coordinates=coordinates
            ).normalize().coordinates

        n_coordinates = np.length(coordinates)

        x = coordinates[:, 0]
        y = coordinates[:, 1]

        LE_index = np.argmin(x)
        is_upper = np.arange(len(x)) <= LE_index

        # Class function
        C = (x) ** N1 * (1 - x) ** N2

        # Shape function (Bernstein polynomials)
        N = n_weights_per_side - 1  # Order of Bernstein polynomials

        K = comb(N, np.arange(N + 1))  # Bernstein polynomial coefficients

        dims = (n_weights_per_side, n_coordinates)

        def wide(vector):
            return np.tile(np.reshape(vector, (1, dims[1])), (dims[0], 1))

        def tall(vector):
            return np.tile(np.reshape(vector, (dims[0], 1)), (1, dims[1]))

        S_matrix = (
                tall(K) * wide(x) ** tall(np.arange(N + 1)) *
                wide(1 - x) ** tall(N - np.arange(N + 1))
        )  # Bernstein polynomial coefficients * weight matrix

        leading_edge_weight_row = x * np.maximum(1 - x, 0) ** (n_weights_per_side + 0.5)

        trailing_edge_thickness_row = np.where(
            is_upper,
            x / 2,
            -x / 2
        )

        A = np.concatenate([
            np.where(wide(is_upper), 0, wide(C) * S_matrix).T,
            np.where(wide(is_upper), wide(C) * S_matrix, 0).T,
            np.reshape(leading_edge_weight_row, (n_coordinates, 1)),
            np.reshape(trailing_edge_thickness_row, (n_coordinates, 1)),
        ], axis=1)

        b = y

        # Solve least-squares problem
        x, _, _, _ = np.linalg.lstsq(A, b, rcond=None)

        lower_weights = x[:n_weights_per_side]
        upper_weights = x[n_weights_per_side:2 * n_weights_per_side]
        leading_edge_weight = x[-2]
        trailing_edge_thickness = x[-1]

        # If you got a negative trailing-edge thickness, then resolve the problem with a TE_thickness = 0 constraint.
        if trailing_edge_thickness < 0:

            x, _, _, _ = np.linalg.lstsq(A[:, :-1], b, rcond=None)

            lower_weights = x[:n_weights_per_side]
            upper_weights = x[n_weights_per_side:2 * n_weights_per_side]
            leading_edge_weight = x[-1]
            trailing_edge_thickness = 0

        return {
            "lower_weights"      : lower_weights,
            "upper_weights"      : upper_weights,
            "TE_thickness"       : trailing_edge_thickness,
            "leading_edge_weight": leading_edge_weight,
        }

    else:
        raise ValueError(f"Invalid method '{method}'.")


def get_coordinates_from_raw_dat(
        raw_text: List[str]
) -> np.ndarray:
    """
    Returns a Nx2 ndarray of airfoil coordinates from the raw text of a airfoil *.dat file.

    Args:

        raw_text: A list of strings, where each string is one line of the *.dat file. One good way to get this input
            is to read the file via the `with open(file, "r") as file:`, `file.readlines()` interface.

    Returns: A Nx2 ndarray of airfoil coordinates [x, y].

    """
    raw_coordinates = []

    def is_number(s: str) -> bool:
        # Determines whether a string is representable as a float
        try:
            float(s)
        except ValueError:
            return False
        return True

    def parse_line(line: str) -> Optional[List[float]]:
        # Given a single line of a `*.dat` file, tries to parse it into a list of two floats [x, y].
        # If not possible, returns None.
        line_split = re.split(r'[;|,|\s|\t]', line)
        line_items = [s for s in line_split if s != ""]
        if len(line_items) == 2 and all([is_number(item) for item in line_items]):
            return line_items
        else:
            return None

    for line in raw_text:
        parsed_line = parse_line(line)
        if parsed_line is not None:
            raw_coordinates.append(parsed_line)

    if len(raw_coordinates) == 0:
        raise ValueError("Could not read any coordinates from the `raw_text` input!")

    coordinates = np.array(raw_coordinates, dtype=float)

    return coordinates


def get_file_coordinates(
        filepath: Union[str, os.PathLike]
):
    possible_errors = (FileNotFoundError, UnicodeDecodeError)

    if isinstance(filepath, np.ndarray):
        raise TypeError("`filepath` should be a string or os.PathLike object.")

    try:
        with open(filepath, "r") as f:
            raw_text = f.readlines()
    except possible_errors as e:
        try:
            with open(f"{filepath}.dat", "r") as f:
                raw_text = f.readlines()
        except possible_errors as e:
            raise FileNotFoundError(
                f" Neither '{filepath}' nor '{filepath}.dat' were found and readable."
            ) from e

    try:
        return get_coordinates_from_raw_dat(raw_text)
    except ValueError:
        raise ValueError("File was found, but could not read any coordinates!")


def get_UIUC_coordinates(
        name: str = 'dae11'
) -> np.ndarray:
    """
    Returns the coordinates of a specified airfoil in the UIUC airfoil database.
    Args:
        name: Name of the airfoil to retrieve from the UIUC database.

    Returns: The coordinates of the airfoil as a Nx2 ndarray [x, y]
    """
    from aerosandbox import _asb_root

    airfoil_database_root = _asb_root / "geometry" / "airfoil" / "airfoil_database"

    try:
        with open(airfoil_database_root / name) as f:
            raw_text = f.readlines()
    except FileNotFoundError as e:
        try:
            with open(airfoil_database_root / f"{name}.dat") as f:
                raw_text = f.readlines()
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Neither '{name}' nor '{name}.dat' were found in the UIUC airfoil database."
            ) from e

    return get_coordinates_from_raw_dat(raw_text)


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np

    af = asb.Airfoil("e377").normalize()
    af.draw(backend="plotly")

    kulfan_params = get_kulfan_parameters(
        coordinates=af.coordinates,
        n_weights_per_side=8,
    )

    af_reconstructed = asb.Airfoil(
        name="Reconstructed Airfoil",
        coordinates=get_kulfan_coordinates(
            **kulfan_params
        ),
    )
    af_reconstructed.draw(backend="plotly")

    print(kulfan_params)
    print(af.jaccard_similarity(af_reconstructed))



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

wing_airfoil = asb.Airfoil("naca0010")  # asb.Airfoil("sd7037")
tail_airfoil = asb.Airfoil("naca0010")

### Define the 3D geometry you want to analyze/optimize.
# Here, all distances are in meters and all angles are in degrees.
airplane = asb.Airplane(
    name="Peter's Glider",
    xyz_ref=[0.18 * 0.32, 0, 0],  # CG location
    s_ref=0.292,
    c_ref=0.151,
    b_ref=2,
    wings=[
        asb.Wing(
            name="Main Wing",
            symmetric=True,  # Should this wing be mirrored across the XZ plane?
            xsecs=[  # The wing's cross ("X") sections
                asb.WingXSec(  # Root
                    xyz_le=[0, 0, 0],  # Coordinates of the XSec's leading edge, relative to the wing's leading edge.
                    chord=0.18,
                    twist=0,  # degrees
                    airfoil=wing_airfoil,  # Airfoils are blended between a given XSec and the next one.
                ),
                asb.WingXSec(  # Mid
                    xyz_le=[0.01, 0.5, 0],
                    chord=0.16,
                    twist=0,
                    airfoil=wing_airfoil,
                ),
                asb.WingXSec(  # Tip
                    xyz_le=[0.08, 1, 0.1],
                    chord=0.08,
                    twist=-0,
                    airfoil=wing_airfoil,
                ),
            ]
        ),
        asb.Wing(
            name="Horizontal Stabilizer",
            symmetric=True,
            xsecs=[
                asb.WingXSec(  # root
                    xyz_le=[0, 0, 0],
                    chord=0.1,
                    twist=-10,
                    airfoil=tail_airfoil,
                ),
                asb.WingXSec(  # tip
                    xyz_le=[0.02, 0.17, 0],
                    chord=0.08,
                    twist=-10,
                    airfoil=tail_airfoil
                )
            ]
        ).translate([0.6, 0, 0.06]),
        asb.Wing(
            name="Vertical Stabilizer",
            symmetric=False,
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=0.1,
                    twist=0,
                    airfoil=tail_airfoil,
                ),
                asb.WingXSec(
                    xyz_le=[0.04, 0, 0.15],
                    chord=0.06,
                    twist=0,
                    airfoil=tail_airfoil
                )
            ]
        ).translate([0.6, 0, 0.07])
    ],
    fuselages=[
        # asb.Fuselage(
        #     name="Fuselage",
        #     xsecs=[
        #         asb.FuselageXSec(
        #             xyz_c=[0.8 * xi - 0.1, 0, 0.1 * xi - 0.03],
        #             radius=0.6 * asb.Airfoil("dae51").local_thickness(x_over_c=xi)
        #         )
        #         for xi in np.cosspace(0, 1, 30)
        #     ]
        # )
    ]
)

op_point = asb.OperatingPoint(
    velocity=100,
    alpha=5,
    beta=0,
)

ab = asb.AeroBuildup(
    airplane,
    op_point,
    xyz_ref=airplane.xyz_ref
).run_with_stability_derivatives()

av = asb.AVL(
    airplane,
    op_point,
    xyz_ref=airplane.xyz_ref
).run()

vl = asb.VortexLatticeMethod(
    airplane,
    op_point,
    xyz_ref=airplane.xyz_ref
).run_with_stability_derivatives()

keys = set()
keys.update(ab.keys())
keys.update(av.keys())
keys = list(keys)
keys.sort()

titles = [
    'Output',
    'AeroBuildup',
    'AVL      ',
    'VLM      ',
    'AB & AVL Significantly Different?'
]


def println(*data):
    print(
        " | ".join([
            d.ljust(len(t)) if isinstance(d, str) else f"{{0:{len(t)}.3g}}".format(d)
            for d, t in zip(data, titles)
        ])
    )


println(*titles)
print("-" * 80)
for k in keys:
    try:
        rel = 0.20
        abs = 0.01

        if 'l' in k or 'm' in k or 'n' in k:
            rel = 0.5
            abs = 0.05

        differences = ab[k] != pytest.approx(av[k], rel=rel, abs=abs)
        differences_text = '*' if differences else ''
        if 'D' in k:
            differences_text = 'Expected'

        println(
            k,
            float(ab[k]),
            float(av[k]),
            float(vl[k]) if k in vl.keys() else ' ' * 5 + '-',
            differences_text
        )

    except (KeyError, TypeError):
        pass



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
from typing import List


def get_airfoil_database() -> List[asb.Airfoil]:
    airfoil_database_root = asb._asb_root / "geometry" / "airfoil" / "airfoil_database"

    afs = [
        asb.Airfoil(
            name=airfoil_file.stem,
        )
        for airfoil_file in airfoil_database_root.glob("*.dat")
    ]

    return afs


def check_validity(af: asb.Airfoil) -> None:
    """
    Checks whether a given Airfoil is valid.

    If the Airfoil is valid, does nothing (returns None).

    If it's egregiously invalid, raises a ValueError.

    If it's valid, but strange enough to warrant human review to double-check it, raises a UserWarning.

    """
    if af.n_points() < 4:
        raise ValueError(f"Airfoil {af.name} has too few points (n_points = {af.n_points()})!")

    if af.area() < 0:
        raise ValueError(f"Airfoil {af.name} has negative area (area = {af.area()})!")

    if af.area() == 0:
        raise ValueError(f"Airfoil {af.name} has zero area!")

    if af.area() > 0.6:
        raise UserWarning(f"Airfoil {af.name} has unusually large area (area = {af.area()})!")

    if af.x().max() > 1.1:
        raise UserWarning(f"Airfoil {af.name} has unusually high x_max (x_max = {af.x().max()})!")

    if af.x().max() < 0.9:
        raise UserWarning(f"Airfoil {af.name} has unusually low x_max (x_max = {af.x().max()})!")

    if af.x().min() < -0.1:
        raise UserWarning(f"Airfoil {af.name} has unusually low x_min (x_min = {af.x().min()})!")

    if af.x().min() > 0.1:
        raise UserWarning(f"Airfoil {af.name} has unusually high x_min (x_min = {af.x().min()})!")

    if af.y().max() > 0.5:
        raise UserWarning(f"Airfoil {af.name} has unusually high y_max (y_max = {af.y().max()})!")

    if af.y().min() < -0.5:
        raise UserWarning(f"Airfoil {af.name} has unusually low y_min (y_min = {af.y().min()})!")

    ## Check for any duplicate points
    ds = np.linalg.norm(np.diff(af.coordinates, axis=0), axis=1)

    if np.any(ds <= 0):
        raise ValueError(f"Airfoil {af.name} has duplicate points (index = {np.argmin(ds)})!")

    if ds.max() > 0.8:
        raise UserWarning(f"Airfoil {af.name} has unusually large ds_max (ds_max = {ds.max()})!")

    ## Check for any negative thickness regions
    x_thicknesses = np.linspace(af.x().min(), af.x().max(), 501)
    thicknesses = af.local_thickness(x_over_c=x_thicknesses)
    if np.any(thicknesses < 0):
        raise ValueError(f"Airfoil {af.name} has negative thickness @ x = {x_thicknesses[np.argmin(thicknesses)]}!")

    ### Make sure the TE thickness is nonnegative
    if af.TE_thickness() < 0:
        raise ValueError(f"Airfoil {af.name} has negative trailing edge thickness {af.TE_thickness()}!")

    ### Make sure the TE angle is not exactly zero, or negative, if the TE thickness is zero.
    if af.TE_thickness() <= 0:
        if af.TE_angle() <= 0:
            raise ValueError(f"Airfoil {af.name} has trailing edge angle {af.TE_angle()}!")

    ### See if Shapely has any complaints
    try:
        import shapely
        if not af.as_shapely_polygon().is_valid:
            raise ValueError(f"Airfoil {af.name} is not a valid Shapely polygon!")

    except ImportError:
        pass


def test_airfoil_database_validity():
    afs = get_airfoil_database()

    failed_airfoils_and_errors = {}

    for af in afs:
        try:
            check_validity(af)
        except UserWarning as e:  # If a UserWarning is raised, print it and continue.
            print(e)
        except ValueError as e:
            failed_airfoils_and_errors[af.name] = e

    if len(failed_airfoils_and_errors) > 0:
        raise ValueError(
            f"The following airfoils failed the validity test:\n"
            "\n".join(f"{af_name}: {error}" for af_name, error in failed_airfoils_and_errors.items())
        )


def debug_draw(af: asb.Airfoil):
    if isinstance(af, str):
        af = asb.Airfoil(af)

    af.draw(
        draw_mcl=False, backend='plotly', show=False
    ).update_layout(
        yaxis=dict(scaleanchor=None)
    ).show()


if __name__ == '__main__':

    afs = get_airfoil_database()
    for af in afs:
        try:
            check_validity(af)
        except (ValueError, UserWarning) as e:
            print(e)
            af.draw()

    # pytest.main()



from aerosandbox.geometry.common import *
from aerosandbox.geometry.airfoil import Airfoil, KulfanAirfoil
from aerosandbox.geometry.wing import Wing, WingXSec, ControlSurface
from aerosandbox.geometry.fuselage import Fuselage, FuselageXSec
from aerosandbox.geometry.airplane import Airplane
from aerosandbox.geometry.propulsor import Propulsor
import aerosandbox.geometry.mesh_utilities as mesh_utils



import aerosandbox.numpy as np


def column_buckling_critical_load(
        elastic_modulus: float,
        moment_of_inertia: float,
        length: float,
        boundary_condition_type: str = "pin-pin",
        use_recommended_design_values: bool = True,
):
    """
    Computes the critical load (in N) for a column or tube in compression to buckle via primary buckling. Uses Euler's classical critical
    load formula.

    Args:
        elastic_modulus: The elastic modulus of the material, in Pa.

        moment_of_inertia: The moment of inertia of the cross-section, in m^4.

        length: The length of the column, in m.

        boundary_condition_type: The boundary condition type. Options are:
            - "pin-pin"
            - "pin-clamp"
            - "clamp-clamp"
            - "clamp-pin"
            - "clamp-free"
            - "free-clamp"

        use_recommended_design_values: Whether to use the recommended design value of K for a given boundary condition (True)
        or to use the less-conservative theoretical value (False).

            * Recommended values are from Table C.1.8.1 in Steel Construction Manual, 8th edition, 2nd revised
            printing, American Institute of Steel Construction, 1987 via WikiMedia:
            https://commons.wikimedia.org/wiki/File:ColumnEffectiveLength.png

    Returns:
        The critical compressive load (in N) for the column or tube to buckle via primary buckling.

    """
    if boundary_condition_type == "pin-pin":
        K = 1.00 if use_recommended_design_values else 1.00
    elif boundary_condition_type == "pin-clamp" or boundary_condition_type == "clamp-pin":
        K = 0.80 if use_recommended_design_values else 0.70
    elif boundary_condition_type == "clamp-clamp":
        K = 0.65 if use_recommended_design_values else 0.50
    elif boundary_condition_type == "clamp-free" or boundary_condition_type == "free-clamp":
        K = 2.10 if use_recommended_design_values else 2.00
    else:
        raise ValueError("Invalid `boundary_condition_type`.")

    return (
            np.pi ** 2 * elastic_modulus * moment_of_inertia
            / (K * length) ** 2
    )


def thin_walled_tube_crippling_buckling_critical_load(
        elastic_modulus: float,
        wall_thickness: float,
        radius: float,
        use_recommended_design_values: bool = True,
):
    """
    Computes the critical load for a thin-walled tube in compression to fail in the crippling mode. (Note: you should also check for
    failure by primary buckling using the `column_buckling_critical_load()` function.)

    The crippling mode is a specific instability mode for tubes with thin walls when loaded in compression. It can be
    seen when you step on a soda can and it buckles inwards. The critical load for this mode is given by the
    following formula:

        stress_crippling = crippling_constant * (E * t / r)

    where:

        A recommended value of crippling_constant = 0.3 is given in Raymer: Aircraft Design: A Conceptual Approach,
        5th Edition, Eq. 14.33, pg. 554.

        A theoretically more accurate value of crippling_constant = 0.605 is given in the Air Force Stress Manual,
        Section 2.3.2.1, Eq. 2-20. This value assumes mu = 0.3, which is a good assumption for most metals.

        and E is the elastic modulus, t is the wall thickness, and r is the radius.

    For more info, see the Air Force Stress Manual, Section 2.3.2.1:
    https://engineeringlibrary.org/reference/column-crippling-air-force-stress-manual

    And see Raymer: Aircraft Design: A Conceptual Approach, 5th Edition, pg. 554.

    Args:
        elastic_modulus: The elastic modulus of the material, in Pa.

        wall_thickness: The wall thickness of the tube, in m.

        radius: The radius of the tube, in m.

        use_recommended_design_values: Whether to use the recommended design value of crippling_constant (True)
        or to use the less-conservative theoretical value (False).

    Returns:
        The critical compressive load (in N) for the tube to buckle in the crippling mode.

    """
    if use_recommended_design_values:
        crippling_stress_constant = 0.3
        # Taken from Raymer: Aircraft Design: A Conceptual Approach, 5th Edition, Eq. 14.33, pg. 554.
        #
        # According to the Air Force Stress Manual, Figure 2-67, this value should drop as radius/wall_thickness
        # increases.
    else:
        crippling_stress_constant = 0.605
        # Theoretically, this should be (3 * (1 - mu^2))^(-0.5), where mu is the Poisson's ratio.
        # Following the Air Force Stress Manual, Section 2.3.2.1, Eq. 2-20.
        # The above value assumes mu = 0.3, which is a good assumption for most metals.

    crippling_stress = 0.3 * (elastic_modulus * wall_thickness / radius)

    tube_xsec_area = 2 * np.pi * radius * wall_thickness

    crippling_load = crippling_stress * tube_xsec_area

    return crippling_load


def plate_buckling_critical_load(
        length: float,
        width: float,
        wall_thickness: float,
        elastic_modulus: float,
        poissons_ratio: float = 0.33,
        side_boundary_condition_type: str = "clamp-clamp",
):
    """
    Computes the critical compressive load (in N) for a plate to buckle via plate buckling.

    Assumes a rectangular plate with dimensions:
    - length
    - width
    - wall_thickness

    A compressive force is applied such that it is aligned with the length dimension of the plate.

    Uses constants from NACA TN3781.

    Methdology taken from "Stress Analysis Manual," Air Force Flight Dynamic Laboratory, Oct. 1986.
    Section 6.3: Axial Compression of Flat Plates
    Reproduced at "Engineering Library":
    https://engineeringlibrary.org/reference/analysis-of-plates-axial-compression-air-force-stress-manual

    Args:
        length: The length of the plate, in m.

        width: The width of the plate, in m.

        wall_thickness: The wall thickness of the plate, in m.

        elastic_modulus: The elastic modulus of the material, in Pa.

        side_boundary_condition_type: The boundary condition type at the sides of the plate. Options are:
            - "clamp-clamp"
            - "pin-pin"
            - "free-free"

    Returns:
        The critical compressive load (in N) for the plate to buckle via plate buckling.

    """

    if side_boundary_condition_type == "clamp-clamp":
        K = 6.35  # From NACA TN3781
    elif side_boundary_condition_type == "pin-pin":
        K = 3.62  # From NACA TN3781
    elif side_boundary_condition_type == "free-free":
        K = 0.385  # From NACA TN3781
    else:
        raise ValueError("Invalid `side_boundary_condition_type`.")

    critical_buckling_load = (
            K * np.pi ** 2 * elastic_modulus /
            (12 * (1 - poissons_ratio ** 2)) *
            wall_thickness ** 3 / width
    )

    return critical_buckling_load



import aerosandbox.numpy as np
from typing import Union
import casadi as cas


def _calculate_induced_velocity_line_singularity_panel_coordinates(
        xp_field: Union[float, np.ndarray],
        yp_field: Union[float, np.ndarray],
        gamma_start: float = 0.,
        gamma_end: float = 0.,
        sigma_start: float = 0.,
        sigma_end: float = 0.,
        xp_panel_end: float = 1.,
) -> [Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point (xp_field, yp_field) in a 2D potential-flow flowfield.

    The `p` suffix in `xp...` and `yp...` denotes the use of the panel coordinate system, where:
        * xp_hat is along the length of the panel
        * yp_hat is orthogonal (90 deg. counterclockwise) to it.

    In this flowfield, there is only one singularity element: A line vortex going from (0, 0) to (xp_panel_end, 0).
    The strength of this vortex varies linearly from:
        * gamma_start at (0, 0), to:
        * gamma_end at (xp_panel_end, 0). # TODO update paragraph

    By convention here, positive gamma induces clockwise swirl in the flow field.
        
    Function returns the 2D velocity u, v in the local coordinate system of the panel.

    Inputs x and y can be 1D ndarrays representing various field points,
    in which case the resulting velocities u and v have corresponding dimensionality.

    Equations from the seminal textbook "Low Speed Aerodynamics" by Katz and Plotkin.
    Vortex equations are Eq. 11.99 and Eq. 11.100.
        * Note: there is an error in equation 11.100 in Katz and Plotkin, at least in the 2nd ed:
        The last term of equation 11.100, which is given as:
            (x_{j+1} - x_j) / z + (theta_{j+1} - theta_j)
        has a sign error and should instead be written as:
            (x_{j+1} - x_j) / z - (theta_{j+1} - theta_j)
    Source equations are Eq. 11.89 and Eq. 11.90.

    """
    ### Modify any incoming floats
    if isinstance(xp_field, (float, int)):
        xp_field = np.array([xp_field])
    if isinstance(yp_field, (float, int)):
        yp_field = np.array([yp_field])

    ### Determine if you can skip either the vortex or source parts
    skip_vortex_math = not (
            isinstance(gamma_start, cas.MX) or
            isinstance(gamma_end, cas.MX)
    ) and gamma_start == 0 and gamma_end == 0
    skip_source_math = not (
            isinstance(sigma_start, cas.MX) or
            isinstance(sigma_end, cas.MX)
    ) and sigma_start == 0 and sigma_end == 0

    ### Determine which points are effectively on the panel, necessitating different math:
    is_on_panel = np.fabs(yp_field) <= 1e-8

    ### Do some geometry calculation
    r_1 = (
                  xp_field ** 2 +
                  yp_field ** 2
          ) ** 0.5
    r_2 = (
                  (xp_field - xp_panel_end) ** 2 +
                  yp_field ** 2
          ) ** 0.5

    ### Regularize
    is_on_endpoint = (
            (r_1 == 0) | (r_2 == 0)
    )
    r_1 = np.where(
        r_1 == 0,
        1,
        r_1,
    )
    r_2 = np.where(
        r_2 == 0,
        1,
        r_2
    )

    ### Continue geometry calculation
    theta_1 = np.arctan2(yp_field, xp_field)
    theta_2 = np.arctan2(yp_field, xp_field - xp_panel_end)
    ln_r_2_r_1 = np.log(r_2 / r_1)
    d_theta = theta_2 - theta_1
    tau = 2 * np.pi

    ### Regularize if the point is on the panel.
    yp_field_regularized = np.where(
        is_on_panel,
        1,
        yp_field
    )

    ### VORTEX MATH
    if skip_vortex_math:
        u_vortex = 0
        v_vortex = 0
    else:
        d_gamma = gamma_end - gamma_start
        u_vortex_term_1_quantity = (yp_field
                                    / tau
                                    * d_gamma
                                    / xp_panel_end
                                    )
        u_vortex_term_2_quantity = (
                                           gamma_start * xp_panel_end + d_gamma * xp_field
                                   ) / (
                                           tau * xp_panel_end
                                   )

        # Calculate u_vortex
        u_vortex_term_1 = u_vortex_term_1_quantity * ln_r_2_r_1
        u_vortex_term_2 = u_vortex_term_2_quantity * d_theta
        u_vortex = u_vortex_term_1 + u_vortex_term_2

        # Correct the u-velocity if field point is on the panel
        u_vortex = np.where(
            is_on_panel,
            0,
            u_vortex
        )

        # Calculate v_vortex
        v_vortex_term_1 = u_vortex_term_2_quantity * ln_r_2_r_1

        v_vortex_term_2 = np.where(
            is_on_panel,
            d_gamma / tau,
            u_vortex_term_1_quantity * (
                    xp_panel_end / yp_field_regularized -
                    d_theta
            ),
        )

        v_vortex = v_vortex_term_1 + v_vortex_term_2

    ### SOURCE MATH
    if skip_source_math:
        u_source = 0
        v_source = 0
    else:
        d_sigma = sigma_end - sigma_start
        v_source_term_1_quantity = (yp_field
                                    / tau
                                    * d_sigma
                                    / xp_panel_end
                                    )
        v_source_term_2_quantity = (
                                           sigma_start * xp_panel_end + d_sigma * xp_field
                                   ) / (
                                           tau * xp_panel_end
                                   )
        # Calculate v_source
        v_source_term_1 = -v_source_term_1_quantity * ln_r_2_r_1
        v_source_term_2 = v_source_term_2_quantity * d_theta
        v_source = v_source_term_1 + v_source_term_2

        # Correct the v-velocity if field point is on the panel
        v_source = np.where(
            is_on_panel,
            0,
            v_source
        )

        # Calculate u_source
        u_source_term_1 = -v_source_term_2_quantity * ln_r_2_r_1

        u_source_term_2 = np.where(
            is_on_panel,
            -d_sigma / tau,
            -v_source_term_1_quantity * (
                    xp_panel_end / yp_field_regularized -
                    d_theta
            ),
        )

        u_source = u_source_term_1 + u_source_term_2

    ### Return
    u = u_vortex + u_source
    v = v_vortex + v_source

    ### If the field point is on the endpoint of the panel, replace the NaN with a zero.
    u = np.where(
        is_on_endpoint,
        0,
        u
    )
    v = np.where(
        is_on_endpoint,
        0,
        v
    )

    return u, v


def _calculate_induced_velocity_line_singularity(
        x_field: Union[float, np.ndarray],
        y_field: Union[float, np.ndarray],
        x_panel_start: float,
        y_panel_start: float,
        x_panel_end: float,
        y_panel_end: float,
        gamma_start: float = 0.,
        gamma_end: float = 0.,
        sigma_start: float = 0.,
        sigma_end: float = 0.,
) -> [Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point (x_field, y_field) in a 2D potential-flow flowfield.

    In this flowfield, there is only one singularity element: # TODO update paragraph
    A line vortex going from (x_panel_start, y_panel_start) to (x_panel_end, y_panel_end).
    The strength of this vortex varies linearly from:
        * gamma_start at (x_panel_start, y_panel_start), to:
        * gamma_end at (x_panel_end, y_panel_end).

    By convention here, positive gamma induces clockwise swirl in the flow field.

    Function returns the 2D velocity u, v in the global coordinate system (x, y).

    Inputs x and y can be 1D ndarrays representing various field points,
    in which case the resulting velocities u and v have the corresponding dimensionality.

    """
    ### Calculate the panel coordinate transform (x -> xp, y -> yp), where
    panel_dx = x_panel_end - x_panel_start
    panel_dy = y_panel_end - y_panel_start
    panel_length = (panel_dx ** 2 + panel_dy ** 2) ** 0.5

    panel_length = np.fmax(panel_length, 1e-16)

    xp_hat_x = panel_dx / panel_length  # x-coordinate of the xp_hat vector
    xp_hat_y = panel_dy / panel_length  # y-coordinate of the yp_hat vector

    yp_hat_x = -xp_hat_y
    yp_hat_y = xp_hat_x

    ### Transform the field points in to panel coordinates
    x_field_relative = x_field - x_panel_start
    y_field_relative = y_field - y_panel_start

    xp_field = x_field_relative * xp_hat_x + y_field_relative * xp_hat_y  # dot product with the xp unit vector
    yp_field = x_field_relative * yp_hat_x + y_field_relative * yp_hat_y  # dot product with the xp unit vector

    ### Do the vortex math
    up, vp = _calculate_induced_velocity_line_singularity_panel_coordinates(
        xp_field=xp_field,
        yp_field=yp_field,
        gamma_start=gamma_start,
        gamma_end=gamma_end,
        sigma_start=sigma_start,
        sigma_end=sigma_end,
        xp_panel_end=panel_length,
    )

    ### Transform the velocities in panel coordinates back to global coordinates
    u = up * xp_hat_x + vp * yp_hat_x
    v = up * xp_hat_y + vp * yp_hat_y

    ### Return
    return u, v


def calculate_induced_velocity_line_singularities(
        x_field: Union[float, np.ndarray],
        y_field: Union[float, np.ndarray],
        x_panels: np.ndarray,
        y_panels: np.ndarray,
        gamma: np.ndarray,
        sigma: np.ndarray,
) -> [Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point (x_field, y_field) in a 2D potential-flow flowfield.

    In this flowfield, the following singularity elements are assumed: # TODO update paragraph
        * A line vortex that passes through the coordinates specified in (x_panel, y_panel). Each of these vertices is
        called a "node".
        * The vorticity of this line vortex per unit length varies linearly between subsequent nodes.
        * The vorticity at each node is specified by the parameter gamma.

    By convention here, positive gamma induces clockwise swirl in the flow field.

    Function returns the 2D velocity u, v in the global coordinate system (x, y).

    Inputs x_field and y_field can be 1D ndarrays representing various field points,
    in which case the resulting velocities u and v have the corresponding dimensionality.

    """
    try:
        N = len(x_panels)
    except TypeError:
        N = x_panels.shape[0]

    u_field = 0.
    v_field = 0.

    for i in range(N - 1):
        u, v = _calculate_induced_velocity_line_singularity(
            x_field=x_field,
            y_field=y_field,
            x_panel_start=x_panels[i],
            y_panel_start=y_panels[i],
            x_panel_end=x_panels[i + 1],
            y_panel_end=y_panels[i + 1],
            gamma_start=gamma[i],
            gamma_end=gamma[i + 1],
            sigma_start=sigma[i],
            sigma_end=sigma[i + 1],
        )
        u_field += u
        v_field += v

    return u_field, v_field


if __name__ == '__main__':

    X, Y = np.meshgrid(
        np.linspace(-2, 2, 50),
        np.linspace(-2, 2, 50),
    )
    X = X.flatten()
    Y = Y.flatten()

    x_panels = np.array([1, -1, -1, 1, 1])
    y_panels = np.array([1, 1, -1, -1, 1])

    U, V = calculate_induced_velocity_line_singularities(
        x_field=X,
        y_field=Y,
        x_panels=x_panels,
        y_panels=y_panels,
        gamma=1 * np.ones_like(x_panels),
        sigma=1 * np.ones_like(x_panels)
    )

    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set(palette=sns.color_palette("husl"))
    fig, ax = plt.subplots(1, 1, figsize=(6, 6), dpi=200)

    plt.quiver(
        X, Y, U, V,
        (U ** 2 + V ** 2) ** 0.5,
        scale=10
    )
    plt.axis("equal")
    plt.xlabel(r"$x$")
    plt.ylabel(r"$z$")
    plt.title(r"Linear-Strength Vortex: Induced Velocity")
    plt.tight_layout()
    # plt.savefig("C:/Users/User/Downloads/temp.svg")
    plt.show()



from pathlib import Path

db = Path(__file__).parent.parent

for file in db.glob("*.dat"):
    with open(file, "r") as f:
        s = f.read()

    s_fixed = s.encode("ascii", "ignore").decode()

    if s != s_fixed:
        with open(file, "w+") as f:
            f.write(s_fixed)


import sympy as s
from sympy import init_printing

init_printing()

# Gets the integral from x2 to x3, looking at the cubic spline interpolant from x1...x4

# Define the symbols
hm, h, hp = s.symbols("hm h hp", real=True)
dfm, df, dfp = s.symbols("dfm df dfp", real=True)

x1, x2, x3, x4 = s.symbols('x1 x2 x3 x4', real=True)
f1, f2, f3, f4 = s.symbols('f1 f2 f3 f4', real=True)

x = s.symbols('x', real=True)


def simplify(expr, maxdepth=10, _depth=0):
    import copy
    original_expr = copy.copy(expr)
    print(f"Depth: {_depth} | Parsimony: {len(str(expr))}")
    maps = {
        f4 - f3: dfp,
        f3 - f2: df,
        f2 - f1: dfm,
        f4 - f2: dfp + df,
        f3 - f1: df + dfm,
        f4 - f1: dfp + df + dfm,
        x4 - x3: hp,
        x3 - x2: h,
        x2 - x1: hm,
        x4 - x2: hp + h,
        x3 - x1: h + hm,
        x4 - x1: hp + h + hm,
    }
    print("hi1")
    # expr = expr.factor(list(maps.keys()))
    print("hi2")
    expr = expr.subs(maps)
    print("hi3")
    # expr = expr.simplify()
    print("hi4")
    if expr != original_expr:
        if _depth < maxdepth:
            expr = simplify(expr, maxdepth=maxdepth, _depth=_depth + 1)
    return expr


f = (
        f1 * (x - x2) * (x - x3) * (x - x4) / ((x1 - x2) * (x1 - x3) * (x1 - x4)) +
        f2 * (x - x1) * (x - x3) * (x - x4) / ((x2 - x1) * (x2 - x3) * (x2 - x4)) +
        f3 * (x - x1) * (x - x2) * (x - x4) / ((x3 - x1) * (x3 - x2) * (x3 - x4)) +
        f4 * (x - x1) * (x - x2) * (x - x3) / ((x4 - x1) * (x4 - x2) * (x4 - x3))
)
f = simplify(f)

dfdx = f.diff(x)
dfdx = simplify(dfdx)

df2dx = f.diff(x, 2)
df2dx = simplify(df2dx)

res = s.integrate(df2dx ** 2, (x, x2, x3))
res = simplify(res.factor([f1, f2, f3, f4]))

parsimony = len(str(res))
print(s.pretty(res, num_columns=100))
print(f"Parsimony: {parsimony}")

#
# q = s.symbols('q')  # Normalized space for a Bernstein basis.
# # Mapping from x-space to q-space has x=x2 -> q=0, x=x3 -> q=1.
#
# q2 = 0
# q3 = 1
# # q1 = q2 - hm / h
# # q4 = q3 + hp / h
# q1, q4 = s.symbols('q1 q4', real=True)
#
# # Define the Bernstein basis polynomials
# b1 = (1 - q) ** 3
# b2 = 3 * q * (1 - q) ** 2
# b3 = 3 * q ** 2 * (1 - q)
# b4 = q ** 3
#
# c1, c2, c3, c4 = s.symbols('c1 c2 c3 c4', real=True)
#
# # Can solve for c2 and c3 exactly
# c1 = f2
# c4 = f3
#
# f = c1 * b1 + c2 * b2 + c3 * b3 + c4 * b4
#
# f1_cubic = f.subs(q, q1)#.simplify()
# f4_cubic = f.subs(q, q4)#.simplify()
#
# factors = [q1, q4]
# # factors = [f1, f2, f3, f4]
#
# # Solve for c2 and c3
# sol = s.solve(
#     [
#         f1_cubic - f1,
#         f4_cubic - f4,
#     ],
#     [
#         c2,
#         c3,
#     ],
# )
# c2 = sol[c2].factor(factors).simplify()
# c3 = sol[c3].factor(factors).simplify()
#
# f = c1 * b1 + c2 * b2 + c3 * b3 + c4 * b4
#
# integral = (c1 + c2 + c3 + c4) / 4 # God I love Bernstein polynomials
# # integral = s.simplify(integral)
#
# integral = integral.factor(factors).simplify()
#
# parsimony = len(str(integral))
# print(s.pretty(integral, num_columns=100))
# print(f"Parsimony: {parsimony}")



from aerosandbox.common import AeroSandboxObject
from abc import abstractmethod
from typing import Union, Dict, List, Tuple
import aerosandbox.numpy as np


class SurrogateModel(AeroSandboxObject):
    """
    A SurrogateModel is effectively a callable; it only has the __call__ method, and all subclasses must explicitly
    overwrite this. The only reason it is not a callable is that you want to be able to save it to disk (via
    pickling) while also having the capability to save associated data (for example, constants associated with a
    particular model, or underlying data).

    If data is used to generate the SurrogateModel, it should be stored as follows:

        * The independent variable(s) should be stored as SurrogateModel.x_data

            * in the general N-dimensional case, x_data should be a dictionary where: keys are variable names and
            values are float/array

            * in the case of a 1-dimensional input (R^1 -> R^1), x-data should be a float/array.

        * The dependent variable should be stored as SurrogateModel.y_data

            * The type of this variable should be a float or an np.ndarray.

    Even if you don't have any real x_data or y_data to add as SurrogateModel.x_data or SurrogateModel.y_data,
        it's recommended (but not required) that you add values here as examples that users can inspect in order to see
        the data types and array shapes required.
    """

    @abstractmethod
    def __init__(self):
        """
        SurrogateModel is an abstract class; you should not instantiate it directly.
        """
        pass

    @abstractmethod  # If you subclass SurrogateModel, you must overwrite __call__ so that it's a callable.
    def __call__(self,
                 x: Union[int, float, np.ndarray, Dict[str, np.ndarray]]
                 ) -> Union[float, np.ndarray]:
        """
        Evaluates the surrogate model at some given input x.

        The input `x` is of the type:
            * in the general N-dimensional case, a dictionary where keys are variable names and values are float/array.
            * in the case of a 1-dimensional input (R^1 -> R^2), a float/array.

        """
        ### Perform basic type checking on x, if x_data exists as a reference.
        try:
            x_data_is_dict = isinstance(self.x_data, dict)
            input_is_dict = isinstance(x, dict)

            if x_data_is_dict and not input_is_dict:
                raise TypeError(
                    f"The input to this model should be a dict with: keys {self.input_names()}, values as float or array.")

            if input_is_dict and not x_data_is_dict:
                raise TypeError("The input to this model should be a float or array.")
        except NameError:  # If x_data does not exist
            pass

    def __repr__(self) -> str:
        input_names = self.input_names()
        if input_names is not None:
            input_description = f"a dict with: keys {input_names}, values as float or array"
        else:
            input_description = f"a float or array"
        return "\n".join([
            f"SurrogateModel(x) [R^{self.input_dimensionality()} -> R^1]",
            f"\tInput: {input_description}",
            f"\tOutput: float or array",
        ])

    def input_dimensionality(self) -> int:
        """
        Returns the number of inputs that should be supplied in x, where x is the input to the SurrogateModel.
        """
        input_names = self.input_names()
        if input_names is not None:
            return len(input_names)
        else:
            return 1

    def input_names(self) -> Union[List, None]:
        """
        If x (the input to this model) is supposed to be a dict, this method returns the keys that should be part of x.

        If x is 1D and simply takes in floats or arrays, or if no x_data exists, returns None.
        """
        try:
            return list(self.x_data.keys())
        except AttributeError:
            return None

    def plot(self, resolution=250):
        import matplotlib.pyplot as plt

        def axis_range(x_data_axis: np.ndarray) -> Tuple[float, float]:
            """
            Given the entries of one axis of the dependent variable, determine a min/max range over which to plot the fit.
            Args:
                x_data_axis: The entries of one axis of the dependent variable, i.e. x_data["x1"].

            Returns: A tuple representing the (min, max) value over which to plot that axis.
            """
            minval = np.min(x_data_axis)
            maxval = np.max(x_data_axis)

            return (minval, maxval)

        if self.input_dimensionality() == 1:

            ### Parse the x_data
            if self.input_names() is not None:
                x_name = self.x_data.keys()[0]
                x_data = self.x_data.values()[0]

                minval, maxval = axis_range(x_data)

                x_fit = {x_name: np.linspace(minval, maxval, resolution)}
                y_fit = self(x_fit)
            else:
                x_name = "x"
                x_data = self.x_data

                minval, maxval = axis_range(x_data)

                x_fit = np.linspace(minval, maxval, resolution)
                y_fit = self(x_fit)

            ### Plot the 2D figure
            fig = plt.figure(dpi=200)
            plt.plot(
                x_data,
                self.y_data,
                ".k",
                label="Data",
            )
            plt.plot(
                x_fit,
                y_fit,
                "-",
                color="#cb3bff",
                label="Fit",
                zorder=4,
            )
            plt.xlabel(x_name)
            plt.ylabel(rf"$f({x_name})$")
            plt.title(r"Fit of FittedModel")
            plt.tight_layout()
            plt.legend()
            plt.show()

        else:
            raise NotImplementedError()



"""
A series of utilities for working with CSV data extracted from WebPlotDigitizer.

https://automeris.io/WebPlotDigitizer/
https://github.com/ankitrohatgi/WebPlotDigitizer

"""
import numpy as np
from typing import Dict
from pathlib import Path
from typing import Union


def string_to_float(s: str) -> float:
    """Converts a string input to a float. If not possible, returns NaN."""
    try:
        return float(s)
    except ValueError:
        return np.nan


def remove_nan_rows(a: np.ndarray) -> np.ndarray:
    """Removes any rows in a 2D ndarray where any of the entries are NaN."""
    nan_rows = np.any(np.isnan(a), axis=1)
    return a[~nan_rows, :]


def read_webplotdigitizer_csv(
        filename: Union[Path, str],
) -> Dict[str, np.ndarray]:
    """
    Reads a CSV file produced by WebPlotDigitizer (https://automeris.io/WebPlotDigitizer/).

    If there's only one data series, produces a Dict with key "data" and value 2D ndarray.

    If there are multiple data series, produces a Dict with keys of the names and values of 2D ndarrays.

    2D ndarrays are sorted by their X-values before being returned.

    Args:
        filename: Filename, as a string or pathlib Path, or equivalent.

    Returns: A dictionary where keys are series names and values are data points.

    """
    delimiter = ","
    with open(filename, "r") as f:
        lines = f.readlines()

    has_titles = np.any([
        np.isnan(string_to_float(s))
        for s in lines[0].split(delimiter)
    ])

    if has_titles:
        titles = lines[0].split(delimiter)[::2]
        first_data_row = 2
    else:
        titles = ["data"]
        first_data_row = 0

    all_data = np.array([
        [string_to_float(item) for item in line.split(delimiter)]
        for line in lines[first_data_row:]
    ], dtype=float)

    output = {}

    for i, title in enumerate(titles):

        series = all_data[:, 2 * i: 2 * i + 2]
        all_nan_rows = np.all(np.isnan(series), axis=1)
        series = series[~all_nan_rows, :]

        sort_order = np.argsort(series[:, 0])

        output[title] = series[sort_order, :]

    return output



import aerosandbox.numpy as np
from aerosandbox.tools import units as u
from aerosandbox.performance.operating_point import OperatingPoint
from typing import Union, Dict


def motor_electric_performance(
        voltage: Union[float, np.ndarray] = None,
        current: Union[float, np.ndarray] = None,
        rpm: Union[float, np.ndarray] = None,
        torque: Union[float, np.ndarray] = None,
        kv: float = 1000.,  # rpm/volt
        resistance: float = 0.1,  # ohms
        no_load_current: float = 0.4  # amps
) -> Dict[str, Union[float, np.ndarray]]:
    """
    A function for predicting the performance of an electric motor.
    
    Performance equations based on Mark Drela's First Order Motor Model:
    http://web.mit.edu/drela/Public/web/qprop/motor1_theory.pdf
    
    Instructions: Input EXACTLY TWO of the following parameters: voltage, current, rpm, torque.
    
    Exception: You cannot supply the combination of current and torque - this makes for an ill-posed problem.
    
    Note that this function is fully vectorized, so arrays can be supplied to any of the inputs.
    
    Args:
        voltage: Voltage across motor terminals [Volts]
        
        current: Current through motor [Amps]
        
        rpm: Motor rotation speed [rpm]
        
        torque: Motor torque [N m]
        
        kv: voltage constant, in rpm/volt
        
        resistance: resistance, in ohms
        
        no_load_current: no-load current, in amps
        
    Returns:
        A dictionary where keys are: 
            "voltage", 
            "current", 
            "rpm", 
            "torque",
            "shaft power", 
            "electrical power", 
            "efficiency"
            "waste heat"

        And values are corresponding quantities in SI units.

        Note that "efficiency" is just (shaft power) / (electrical power), and hence implicitly assumes that the
        motor is operating as a motor (electrical -> shaft power), and not a generator (shaft power -> electrical).
        If you want to know the efficiency of the motor as a generator, you can simply calculate it as (electrical
        power) / (shaft power).
    """
    # Validate inputs
    voltage_known = voltage is not None
    current_known = current is not None
    rpm_known = rpm is not None
    torque_known = torque is not None

    if not (
            voltage_known + current_known + rpm_known + torque_known == 2
    ):
        raise ValueError("You must give exactly two input arguments.")

    if current_known and torque_known:
        raise ValueError(
            "You cannot supply the combination of current and torque - this makes for an ill-posed problem.")

    kv_rads_per_sec_per_volt = kv * np.pi / 30  # rads/sec/volt

    ### Iterate through the motor equations until all quantities are known.
    while not (voltage_known and current_known and rpm_known and torque_known):
        if rpm_known:
            if current_known and not voltage_known:
                speed = rpm * np.pi / 30  # rad/sec
                back_EMF_voltage = speed / kv_rads_per_sec_per_volt
                voltage = back_EMF_voltage + current * resistance
                voltage_known = True

        if torque_known:
            if not current_known:
                current = torque * kv_rads_per_sec_per_volt + no_load_current
                current_known = True

        if voltage_known:
            if rpm_known and not current_known:
                speed = rpm * np.pi / 30  # rad/sec
                back_EMF_voltage = speed / kv_rads_per_sec_per_volt
                current = (voltage - back_EMF_voltage) / resistance
                current_known = True
            if not rpm_known and current_known:
                back_EMF_voltage = voltage - (current * resistance)
                speed = back_EMF_voltage * kv_rads_per_sec_per_volt
                rpm = speed * 30 / np.pi
                rpm_known = True

        if current_known:
            if not torque_known:
                torque = (current - no_load_current) / kv_rads_per_sec_per_volt
                torque_known = True

    shaft_power = (rpm * np.pi / 30) * torque
    electrical_power = voltage * current
    efficiency = shaft_power / electrical_power
    waste_heat = np.fabs(electrical_power - shaft_power)

    return {
        "voltage"         : voltage,
        "current"         : current,
        "rpm"             : rpm,
        "torque"          : torque,
        "shaft power"     : shaft_power,
        "electrical power": electrical_power,
        "efficiency"      : efficiency,
        "waste heat"      : waste_heat,
    }


def electric_propeller_propulsion_analysis(
        total_thrust: float,
        n_engines: int,
        propeller_diameter: float,
        op_point: OperatingPoint,
        motor_kv: float,
        motor_no_load_current: float,
        motor_resistance: float,
        wire_resistance: float,
        battery_voltage: float,
        propeller_tip_mach: float = 0.50,
        gearbox_ratio: float = 1,
        gearbox_efficiency: float = 1,
        esc_efficiency: float = 0.98,
        battery_discharge_efficiency: float = 0.985,
) -> Dict[str, float]:
    """
    Performs a propulsion analysis for an electric propeller-driven aircraft.

    May be used for single-engine or multi-engine aircraft, so long as all engines / propellers are identical.

    Args:

        total_thrust: Total thrust force produced by all engines at the cruise operating point [N].

        n_engines: Number of engines on the aircraft.

        propeller_diameter: Diameter of each of the propellers [m].

        op_point: The cruise operating point. Must be an AeroSandbox OperatingPoint object.

        motor_kv: Motor voltage constant [rpm/volt].

        motor_no_load_current: Motor no-load current [amps].

        motor_resistance: Motor resistance [ohms].

        wire_resistance: Round-trip resistance of the wires connecting the ESC to the battery [ohms].

        battery_voltage: Battery voltage [volts].

        propeller_tip_mach: Mach number at the propeller tip. Defaults to 0.50. From a propulsive efficiency
            perspective, you want this to be as high as possible while still keeping the tip speed (hypotenuse of the
            velocity triangle) below the critical Mach number of the propeller blade airfoil. This is because motor
            efficiency and specific power tend to be better at high-speed low-torque conditions, and also the propeller
            aerodynamics tend to be better at low solidity. But there may be reasons to lower this, such as propeller
            structural considerations or noise considerations (with noise being a *strong* function of tip Mach).

        gearbox_ratio: Gearbox reduction ratio. Defaults to 1 (no gearbox). For example, a `gearbox_ratio` of 5 is a 5:1
            reduction, meaning that the propeller turns 5 times slower than the motor.

        gearbox_efficiency: Gearbox efficiency. Defaults to 1, only because the `gearbox_ratio` defaults to 1 (no
            gearbox), and so this represents no losses. If you have a gearbox, you should probably use a value of 0.98 or
            so.

        esc_efficiency: Efficiency of the electronic speed controller (ESC), sometimes called the inverter. Defaults to
            0.98, which is a reasonable value for a high-quality ESC at a large (>5 kW) scale. Small components will
            lower efficiencies than this.

        battery_discharge_efficiency: Coulobmic efficiency of the battery in discharge only (i.e., not round-trip).
            Defaults to 0.985, which is a reasonable value for a high-quality lithium-polymer battery. Other battery
            chemistries will have different values.

    Returns: A dictionary of various parameters of the propulsion analysis. Of particular note are the following keys:

        * "air_power": The power delivered to the air (thrust * velocity) [W]
        * "shaft_power": The power at the propeller shaft (after the gearbox; rotational speed * torque) [W]
        * "motor_electrical_power": The electrical power input to the motor [W]
        * "esc_electrical_power": The electrical power input to the ESC [W]
        * "battery_power": The power draw from the battery [W].

        * "propeller_efficiency": The propulsive efficiency of the propeller, defined as (air_power / shaft_power).
        * "motor_efficiency": The efficiency of the motor, defined as (shaft_power / motor_electrical_power).
        * "overall_efficiency": The overall efficiency of the propulsion system, defined as (air_power / battery_power).

    """

    ### Propeller Analysis
    propulsive_area_per_propeller = (np.pi / 4) * propeller_diameter ** 2
    propulsive_area_total = propulsive_area_per_propeller * n_engines

    propeller_wake_dynamic_pressure = op_point.dynamic_pressure() + total_thrust / propulsive_area_total
    propeller_wake_velocity = (
                                  # Derived from the above pressure jump relation, with adjustments to avoid singularity at zero velocity
                                      2 * total_thrust / (propulsive_area_total * op_point.atmosphere.density())
                                      + op_point.velocity ** 2
                              ) ** 0.5

    propeller_tip_speed = propeller_tip_mach * op_point.atmosphere.speed_of_sound()
    propeller_rads_per_sec = propeller_tip_speed / (propeller_diameter / 2)
    propeller_rpm = propeller_rads_per_sec * 60 / (2 * np.pi)

    propeller_advance_ratio = op_point.velocity / propeller_tip_speed

    air_power = total_thrust * op_point.velocity

    from aerosandbox.library.propulsion_propeller import propeller_shaft_power_from_thrust

    shaft_power = propeller_shaft_power_from_thrust(
        thrust_force=total_thrust,
        area_propulsive=propulsive_area_total,
        airspeed=op_point.velocity,
        rho=op_point.atmosphere.density(),
        propeller_coefficient_of_performance=0.90,
    )

    propeller_efficiency = air_power / shaft_power

    ### Motor Analysis
    motor_rpm = propeller_rpm / gearbox_ratio
    motor_rads_per_sec = motor_rpm * 2 * np.pi / 60

    motor_torque_per_motor = shaft_power / n_engines / motor_rads_per_sec / gearbox_efficiency

    motor_parameters_per_motor = motor_electric_performance(
        rpm=motor_rpm,
        torque=motor_torque_per_motor,
        kv=motor_kv,
        no_load_current=motor_no_load_current,
        resistance=motor_resistance,
    )

    motor_electrical_power = motor_parameters_per_motor["electrical power"] * n_engines

    motor_efficiency = shaft_power / motor_electrical_power

    ### ESC Analysis
    esc_electrical_power = motor_electrical_power / esc_efficiency

    ### Wire Analysis
    wire_power_loss = (esc_electrical_power / battery_voltage) ** 2 * wire_resistance
    wire_efficiency = esc_electrical_power / (esc_electrical_power + wire_power_loss)

    ### Battery Analysis
    battery_power = (esc_electrical_power + wire_power_loss) / battery_discharge_efficiency
    battery_current = battery_power / battery_voltage

    ### Overall
    overall_efficiency = air_power / battery_power

    return locals()


def motor_resistance_from_no_load_current(
        no_load_current
):
    """
    Estimates the internal resistance of a motor from its no_load_current. Gates quotes R^2=0.93 for this model.

    Source: Gates, et. al., "Combined Trajectory, Propulsion, and Battery Mass Optimization for Solar-Regen..."
        https://scholarsarchive.byu.edu/cgi/viewcontent.cgi?article=3932&context=facpub

    Args:
        no_load_current: No-load current [amps]

    Returns:
        motor internal resistance [ohms]
    """
    return 0.0467 * no_load_current ** -1.892


def mass_ESC(
        max_power,
):
    """
    Estimates the mass of an ESC.

    Informal correlation I did to Hobbyking ESCs in the 8S LiPo, 100A range

    Args:
        max_power: maximum power [W]

    Returns:
        estimated ESC mass [kg]
    """
    return 2.38e-5 * max_power


def mass_battery_pack(
        battery_capacity_Wh,
        battery_cell_specific_energy_Wh_kg=240,
        battery_pack_cell_fraction=0.7,
):
    """
    Estimates the mass of a lithium-polymer battery.

    Args:
        battery_capacity_Wh: Battery capacity, in Watt-hours [W*h]

        battery_cell_specific_energy: Specific energy of the battery at the CELL level [W*h/kg]

        battery_pack_cell_fraction: Fraction of the battery pack that is cells, by weight.

            * Note: Ed Lovelace, a battery engineer for Aurora Flight Sciences, gives this figure as 0.70 in a Feb.
            2020 presentation for MIT 16.82

    Returns:
        Estimated battery mass [kg]
    """
    return battery_capacity_Wh / battery_cell_specific_energy_Wh_kg / battery_pack_cell_fraction


def mass_motor_electric(
        max_power,
        kv_rpm_volt=1000,  # This is in rpm/volt, not rads/sec/volt!
        voltage=20,
        method="hobbyking"
):
    """
    Estimates the mass of a brushless DC electric motor.
    Curve fit to scraped Hobbyking BLDC motor data as of 2/24/2020.
    Estimated range of validity: 50 < max_power < 10000

    Args:
        max_power (float): maximum power [W]

        kv_rpm_volt (float): Voltage constant of the motor, measured in rpm/volt, not rads/sec/volt! [rpm/volt]

        voltage (float): Operating voltage of the motor [V]

        method (str): method to use. "burton", "hobbyking", or "astroflight" (increasing level of detail).

            * Burton source: https://dspace.mit.edu/handle/1721.1/112414

            * Hobbyking source: C:\Projects\GitHub\MotorScraper,

            * Astroflight source: Gates, et. al., "Combined Trajectory, Propulsion, and Battery Mass Optimization for
            Solar-Regen..." https://scholarsarchive.byu.edu/cgi/viewcontent.cgi?article=3932&context=facpub

                * Validity claimed from 1.5 kW to 15 kW, kv from 32 to 1355.

    Returns:
        Estimated motor mass [kg]
    """
    if method == "burton":
        return max_power / 4128  # Less sophisticated model. 95% CI (3992, 4263), R^2 = 0.866
    elif method == "hobbyking":
        return 10 ** (0.8205 * np.log10(max_power) - 3.155)  # More sophisticated model
    elif method == "astroflight":
        max_current = max_power / voltage
        return 2.464 * max_current / kv_rpm_volt + 0.368  # Even more sophisticated model


def mass_wires(
        wire_length,
        max_current,
        allowable_voltage_drop,
        material="aluminum",
        insulated=True,
        max_voltage=600,
        wire_packing_factor=1,
        insulator_density=1700,
        insulator_dielectric_strength=12e6,
        insulator_min_thickness=0.2e-3,  # silicone wire
        return_dict: bool = False
):
    """
    Estimates the mass of wires used for power transmission.

    Materials data from: https://en.wikipedia.org/wiki/Electrical_resistivity_and_conductivity#Resistivity-density_product
        All data measured at STP; beware, as this data (especially resistivity) can be a strong function of temperature.

    Args:
        wire_length (float): Length of the wire [m]

        max_current (float): Max current of the wire [Amps]

        allowable_voltage_drop (float): How much is the voltage allowed to drop along the wire?

        material (str): Conductive material of the wire ("aluminum"). Determines density and resistivity. One of:

            * "sodium"

            * "lithium"

            * "calcium"

            * "potassium"

            * "beryllium"

            * "aluminum"

            * "magnesium"

            * "copper"

            * "silver"

            * "gold"

            * "iron"

        insulated (bool): Should we add the mass of the wire's insulator coating? Usually you'll want to leave this True.

        max_voltage (float): Maximum allowable voltage (used for sizing insulator). 600 is a common off-the-shelf rating.

        wire_packing_factor (float): What fraction of the enclosed cross section is conductor? This is 1 for solid wire,
            and less for stranded wire.

        insulator_density (float): Density of the insulator [kg/m^3]

        insulator_dielectric_strength (float): Dielectric strength of the insulator [V/m]. The default value of 12e6 corresponds
        to rubber.

        insulator_min_thickness (float): Minimum thickness of the insulator [m]. This is essentially a gauge limit.
        The default value is 0.2 mm.

        return_dict (bool): If True, returns a dictionary of all local variables. If False, just returns the wire
        mass as a float. Defaults to False.


    Returns: If `return_dict` is False (default), returns the wire mass as a single number. If `return_dict` is True,
    returns a dictionary of all local variables.
    """
    if material == "sodium":  # highly reactive with water & oxygen, low physical strength
        density = 970  # kg/m^3
        resistivity = 47.7e-9  # ohm-meters
    elif material == "lithium":  # highly reactive with water & oxygen, low physical strength
        density = 530  # kg/m^3
        resistivity = 92.8e-9  # ohm-meters
    elif material == "calcium":  # highly reactive with water & oxygen, low physical strength
        density = 1550  # kg/m^3
        resistivity = 33.6e-9  # ohm-meters
    elif material == "potassium":  # highly reactive with water & oxygen, low physical strength
        density = 890  # kg/m^3
        resistivity = 72.0e-9  # ohm-meters
    elif material == "beryllium":  # toxic, brittle
        density = 1850  # kg/m^3
        resistivity = 35.6e-9  # ohm-meters
    elif material == "aluminum":
        density = 2700  # kg/m^3
        resistivity = 26.50e-9  # ohm-meters
    elif material == "magnesium":  # worse specific conductivity than aluminum
        density = 1740  # kg/m^3
        resistivity = 43.90e-9  # ohm-meters
    elif material == "copper":  # worse specific conductivity than aluminum, moderately expensive
        density = 8960  # kg/m^3
        resistivity = 16.78e-9  # ohm-meters
    elif material == "silver":  # worse specific conductivity than aluminum, expensive
        density = 10490  # kg/m^3
        resistivity = 15.87e-9  # ohm-meters
    elif material == "gold":  # worse specific conductivity than aluminum, very expensive
        density = 19300  # kg/m^3
        resistivity = 22.14e-9  # ohm-meters
    elif material == "iron":  # worse specific conductivity than aluminum
        density = 7874  # kg/m^3
        resistivity = 96.1e-9  # ohm-meters
    else:
        raise ValueError("Bad value of 'material'!")

    # Conductor mass
    resistance = allowable_voltage_drop / max_current
    area_conductor = resistivity * wire_length / resistance
    volume_conductor = area_conductor * wire_length
    mass_conductor = volume_conductor * density

    # Insulator mass
    if insulated:
        insulator_thickness = np.softmax(
            4.0 * max_voltage / insulator_dielectric_strength,
            insulator_min_thickness,
            softness=0.005 * u.inch,
        )
        radius_conductor = (area_conductor / wire_packing_factor / np.pi) ** 0.5
        radius_insulator = radius_conductor + insulator_thickness
        area_insulator = np.pi * radius_insulator ** 2 - area_conductor
        volume_insulator = area_insulator * wire_length
        mass_insulator = insulator_density * volume_insulator
    else:
        mass_insulator = 0

    # Total them up
    mass_total = mass_conductor + mass_insulator

    if return_dict:
        return locals()
    else:
        return mass_total


if __name__ == '__main__':
    print(motor_electric_performance(
        rpm=100,
        current=3
    ))
    print(motor_electric_performance(
        rpm=4700,
        torque=0.02482817
    ))

    print(
        mass_battery_pack(100)
    )

    pows = np.logspace(2, 5, 300)
    mass_mot_burton = mass_motor_electric(pows, method="burton")
    mass_mot_hobbyking = mass_motor_electric(pows, method="hobbyking")
    mass_mot_astroflight = mass_motor_electric(pows, method="astroflight")

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    plt.loglog(pows, np.array(mass_mot_burton), "-", label="Burton Model")
    plt.plot(pows, np.array(mass_mot_hobbyking), "--", label="Hobbyking Model")
    plt.plot(pows, np.array(mass_mot_astroflight), "-.", label="Astroflight Model")
    p.show_plot(
        "Small Electric Motor Mass Models\n(500 kv, 100 V)",
        "Motor Power [W]",
        "Motor Mass [kg]"
    )

    print(mass_wires(
        wire_length=1,
        max_current=100,
        allowable_voltage_drop=1,
        material="aluminum"
    ))



import pytest
import aerosandbox.numpy as np
import casadi as cas


def test_cross_1D_input():
    a = np.array([1, 1, 1])
    b = np.array([1, 2, 3])

    cas_a = cas.DM(a)
    cas_b = cas.DM(b)

    correct_result = np.cross(a, b)
    cas_correct_result = cas.DM(correct_result)

    assert np.all(
        np.cross(a, cas_b) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, b) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, cas_b) == cas_correct_result
    )


def test_cross_2D_input_last_axis():
    a = np.tile(np.array([1, 1, 1]), (3, 1))
    b = np.tile(np.array([1, 2, 3]), (3, 1))

    cas_a = cas.DM(a)
    cas_b = cas.DM(b)

    correct_result = np.cross(a, b)
    cas_correct_result = cas.DM(correct_result)

    assert np.all(
        np.cross(a, cas_b) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, b) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, cas_b) == cas_correct_result
    )


def test_cross_2D_input_first_axis():
    a = np.tile(np.array([1, 1, 1]), (3, 1)).T
    b = np.tile(np.array([1, 2, 3]), (3, 1)).T

    cas_a = cas.DM(a)
    cas_b = cas.DM(b)

    correct_result = np.cross(a, b, axis=0)
    cas_correct_result = cas.DM(correct_result)

    assert np.all(
        np.cross(a, cas_b, axis=0) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, b, axis=0) == cas_correct_result
    )
    assert np.all(
        np.cross(cas_a, cas_b, axis=0) == cas_correct_result
    )


if __name__ == '__main__':
    pytest.main()



# From Raymer, Aircraft Design: A Conceptual Approach, 5th Ed.
# Table 15.4
advanced_composites = {
    # Format:
    # "component"         : (low, high)
    "wing"                : (0.85, 0.90),
    "tails"               : (0.83, 0.88),
    "fuselage/nacelle"    : (0.90, 0.95),
    "landing_gear"        : (0.95, 1),
    "air_induction_system": (0.85, 0.90),
}

advanced_composites = { # Here, we convert this to a dictionary of average values.
    k: (v[0] + v[1]) / 2
    for k, v in advanced_composites.items()
}

braced_wing = 0.82

braced_biplane = 0.6

wood_fuselage = 1.60

steel_tube_fuselage = 1.80

flying_boat_hull = 1.25

carrier_based_fuselage = 1.25
carrier_based_landing_gear = 1.25
